# cython: embedsignature=True
# cython: freethreading_compatible=True
# cython: language_level=3
from cpython.dict cimport (PyDict_Check, PyDict_CheckExact, PyDict_GetItem,
						   PyDict_New, PyDict_Next,
						   PyDict_SetItem, PyDict_Update, PyDict_DelItem)
from cpython.list cimport PyList_Append, PyList_New
from cpython.object cimport PyObject_SetItem
from cpython.ref cimport PyObject, Py_DECREF, Py_INCREF, Py_XDECREF

# Locally defined bindings that differ from `cython.cpython` bindings
from humpy_cytoolz.cpython cimport PyDict_Next_Compat, PtrIter_Next

from collections import abc

# cdef aliases to eliminate global lookups

cdef object Mapping = abc.Mapping
del abc

__all__ = ['merge', 'merge_with', 'valmap', 'keymap', 'itemmap', 'valfilter',
		   'keyfilter', 'itemfilter', 'assoc', 'dissoc', 'assoc_in', 'get_in',
		   'update_in']

cdef class _iter_mapping:
	""" Keep a handle on the current item to prevent memory clean up too early"""
	def __cinit__(self, object it):
		self.it = it
		self.cur = None

	def __iter__(self):
		return self

	def __next__(self):
		self.cur = next(self.it)
		return self.cur

cdef int PyMapping_Next(object p, Py_ssize_t *ppos, PyObject* *pkey, PyObject* *pval) except -1:
	"""Mimic "PyDict_Next" interface, but for any mapping"""
	cdef PyObject *obj
	obj = PtrIter_Next(p)
	if obj is NULL:
		return 0
	pkey[0] = <PyObject*>(<object>obj)[0]
	pval[0] = <PyObject*>(<object>obj)[1]
	Py_XDECREF(obj)  # removing this results in memory leak
	return 1

cdef f_map_next get_map_iter(object d, PyObject* *ptr) except NULL:
	"""Return function pointer to perform iteration over object returned in ptr.

	The returned function signature matches "PyDict_Next".  If ``d`` is a dict,
	then the returned function *is* PyDict_Next, so iteration wil be very fast.

	The object returned through ``ptr`` needs to have its reference count
	reduced by one once the caller "owns" the object.

	This function lets us control exactly how iteration should be performed
	over a given mapping.  The current rules are:

	1) If ``d`` is exactly a dict, use PyDict_Next
	2) If ``d`` is subtype of dict, use PyMapping_Next.  This lets the user
	   control the order iteration, such as for ordereddict.
	3) If using PyMapping_Next, iterate using ``iteritems`` if possible,
	   otherwise iterate using ``items``.

	"""
	cdef object val
	cdef f_map_next rv
	if PyDict_CheckExact(d):
		val = d
		rv = &PyDict_Next_Compat
	else:
		val = _iter_mapping(iter(d.items()))
		rv = &PyMapping_Next
	Py_INCREF(val)
	ptr[0] = <PyObject*>val
	return rv

cdef get_factory(name, kwargs):
	factory = kwargs.pop('factory', dict)
	if kwargs:
		raise TypeError("{0}() got an unexpected keyword argument "
						"'{1}'".format(name, kwargs.popitem()[0]))
	return factory

cdef object c_merge(object dicts, object factory=dict):
	cdef object rv
	rv = factory()
	if PyDict_CheckExact(rv):
		for d in dicts:
			PyDict_Update(rv, d)
	else:
		for d in dicts:
			rv.update(d)
	return rv

def merge(*dicts, factory=dict):
	"""\
merge(*dicts: collections.abc.Mapping[K, V], factory: collections.abc.Callable[[], collections.abc.MutableMapping[K, V]] = dict) -> collections.abc.MutableMapping[K, V]

Merge a collection of dictionaries and return a new `Mapping`.

(AI generated docstring)

You can use `merge` to combine two or more `Mapping`[1] objects into a single new `Mapping`.
`merge` calls `factory` to create the result, then updates it in order with each `Mapping` in
`dicts`. When the same key appears in more than one element, the value from the later element
takes precedence. You can also pass a single `Iterable[Mapping[K, V]]` as the sole positional
argument instead of multiple `Mapping` arguments.

Parameters
----------
*dicts : Mapping[K, V]
	`Mapping` objects to merge. Alternatively, pass a single `Iterable[Mapping[K, V]]` as the
	sole positional argument.
factory : Callable[[], MutableMapping[K, V]] = dict
	`Callable` that creates the `MutableMapping`[1] to `return`.

Returns
-------
mappingMerged : MutableMapping[K, V]
	New `MutableMapping` created by `factory` containing all key-value pairs from `dicts`. For
	duplicate keys, the value from the last `Mapping` in `dicts` that contains the key takes
	precedence.

See Also
--------
merge_with : Merge dictionaries and apply a `Callable` to combined values.

Examples
--------
>>> merge({1: 'one'}, {2: 'two'})
{1: 'one', 2: 'two'}

Later dictionaries have precedence

>>> merge({1: 2, 3: 4}, {3: 3, 4: 4})
{1: 2, 3: 3, 4: 4}

References
----------
[1] Python `collections.abc` module
	https://docs.python.org/3/library/collections.abc.html
	"""
	if len(dicts) == 1 and not isinstance(dicts[0], Mapping):
		dicts = dicts[0]
	return c_merge(dicts, factory)

cdef object c_merge_with(object func, object dicts, object factory=dict):
	cdef:
		dict result
		object rv, d
		list seq
		f_map_next f
		PyObject *obj
		PyObject *pkey
		PyObject *pval
		Py_ssize_t pos

	result = PyDict_New()
	rv = factory()
	for d in dicts:
		f = get_map_iter(d, &obj)
		d = <object>obj
		Py_DECREF(d)
		pos = 0
		while f(d, &pos, &pkey, &pval):
			obj = PyDict_GetItem(result, <object>pkey)
			if obj is NULL:
				seq = PyList_New(0)
				PyList_Append(seq, <object>pval)
				PyDict_SetItem(result, <object>pkey, seq)
			else:
				PyList_Append(<object>obj, <object>pval)

	f = get_map_iter(result, &obj)
	d = <object>obj
	Py_DECREF(d)
	pos = 0
	while f(d, &pos, &pkey, &pval):
		PyObject_SetItem(rv, <object>pkey, func(<object>pval))
	return rv

def merge_with(func, *dicts, factory=dict):
	"""\
merge_with(func: collections.abc.Callable[[list[V]], V], *dicts: collections.abc.Mapping[K, V], factory: collections.abc.Callable[[], collections.abc.MutableMapping[K, V]] = dict) -> collections.abc.MutableMapping[K, V]

Merge dictionaries and apply a `Callable` to combined values.

(AI generated docstring)

You can use `merge_with` to combine two or more `Mapping`[1] objects and resolve key conflicts
by applying `func` to a `list` of all values associated with each key. For each key that
appears in one or more elements of `dicts`, `merge_with` collects all associated values into a
`list` in the order they appear across `dicts`, then calls `func` with that `list` to produce
the value in the result. You can also pass a single `Iterable[Mapping[K, V]]` as the sole
positional argument after `func`.

Parameters
----------
func : Callable[[list[V]], V]
	`Callable` applied to the `list` of values associated with each key across all `Mapping`
	objects in `dicts`. `func` receives a non-empty `list` and must return the merged value for
	that key.
*dicts : Mapping[K, V]
	`Mapping` objects to merge. Alternatively, pass a single `Iterable[Mapping[K, V]]` as the
	sole positional argument after `func`.
factory : Callable[[], MutableMapping[K, V]] = dict
	`Callable` that creates the `MutableMapping`[1] to `return`.

Returns
-------
mappingMerged : MutableMapping[K, V]
	New `MutableMapping` created by `factory` where each key maps to the result of calling
	`func` with all values associated with that key across `dicts`.

See Also
--------
merge : Merge a collection of `Mapping` objects into a new `Mapping`.

Examples
--------
>>> merge_with(sum, {1: 1, 2: 2}, {1: 10, 2: 20})
{1: 11, 2: 22}

>>> merge_with(first, {1: 1, 2: 2}, {2: 20, 3: 30})  # doctest: +SKIP
{1: 1, 2: 2, 3: 30}

References
----------
[1] Python `collections.abc` module
	https://docs.python.org/3/library/collections.abc.html
	"""
	if len(dicts) == 1 and not isinstance(dicts[0], Mapping):
		dicts = dicts[0]
	return c_merge_with(func, dicts, factory)

cpdef object valmap(object func, object d, object factory=dict):
	"""valmap(func: collections.abc.Callable[[V], W], d: collections.abc.Mapping[K, V], factory: collections.abc.Callable[[], collections.abc.MutableMapping[K, W]] = dict) -> collections.abc.MutableMapping[K, W]

	Apply `func` to all values of `d` and return a new `Mapping` with the transformed values.

	(AI generated docstring)

	You can use `valmap` to transform all values in `d` without changing `d`. `valmap` applies
	`func` to each value yielded by `d.values()` and builds a new `MutableMapping`[1] created
	by `factory`, preserving each original key associated with its transformed value.

	Parameters
	----------
	func : Callable[[V], W]
		`Callable` applied to each value of `d`. Each value of `d` is passed to `func`
		individually, and `func` returns the corresponding transformed value.
	d : Mapping[K, V]
		Source `Mapping`[1]. `valmap` reads all values from `d` and does not change `d`.
	factory : Callable[[], MutableMapping[K, W]] = dict
		`Callable` that creates the `MutableMapping`[1] to `return`.

	Returns
	-------
	mappingTransformed : MutableMapping[K, W]
		New `MutableMapping` created by `factory` in which each key from `d` is associated
		with the result of applying `func` to the corresponding value of `d`.

	See Also
	--------
	keymap : Apply a `Callable` to all keys of a `Mapping` and return a new `Mapping`.
	itemmap : Apply a `Callable` to all items of a `Mapping` and return a new `Mapping`.

	Examples
	--------
	>>> bills = {'Alice': [20, 15, 30], 'Bob': [10, 35]}
	>>> valmap(sum, bills)  # doctest: +SKIP
	{'Alice': 65, 'Bob': 45}

	References
	----------
	[1] Python `collections.abc` module
		https://docs.python.org/3/library/collections.abc.html
	"""
	cdef:
		object rv
		f_map_next f
		PyObject *obj
		PyObject *pkey
		PyObject *pval
		Py_ssize_t pos = 0

	rv = factory()
	f = get_map_iter(d, &obj)
	d = <object>obj
	Py_DECREF(d)
	while f(d, &pos, &pkey, &pval):
		rv[<object>pkey] = func(<object>pval)
	return rv

cpdef object keymap(object func, object d, object factory=dict):
	"""keymap(func: collections.abc.Callable[[K], L], d: collections.abc.Mapping[K, V], factory: collections.abc.Callable[[], collections.abc.MutableMapping[L, V]] = dict) -> collections.abc.MutableMapping[L, V]

	Apply `func` to all keys of `d` and return a new `Mapping` with the transformed keys.

	(AI generated docstring)

	You can use `keymap` to transform all keys in `d` without changing `d`. `keymap` applies
	`func` to each key yielded by `d.keys()` and builds a new `MutableMapping`[1] created by
	`factory`, associating each transformed key with the corresponding original value of `d`.

	Parameters
	----------
	func : Callable[[K], L]
		`Callable` applied to each key of `d`. Each key of `d` is passed to `func`
		individually, and `func` returns the corresponding transformed key.
	d : Mapping[K, V]
		Source `Mapping`[1]. `keymap` reads all keys from `d` and does not change `d`.
	factory : Callable[[], MutableMapping[L, V]] = dict
		`Callable` that creates the `MutableMapping`[1] to `return`.

	Returns
	-------
	mappingTransformed : MutableMapping[L, V]
		New `MutableMapping` created by `factory` in which each key from `d` is replaced by
		the result of applying `func` to that key, associated with the original value of `d`.

	See Also
	--------
	valmap : Apply a `Callable` to all values of a `Mapping` and return a new `Mapping`.
	itemmap : Apply a `Callable` to all items of a `Mapping` and return a new `Mapping`.

	Examples
	--------
	>>> bills = {'Alice': [20, 15, 30], 'Bob': [10, 35]}
	>>> keymap(str.lower, bills)  # doctest: +SKIP
	{'alice': [20, 15, 30], 'bob': [10, 35]}

	References
	----------
	[1] Python `collections.abc` module
		https://docs.python.org/3/library/collections.abc.html
	"""
	cdef:
		object rv
		f_map_next f
		PyObject *obj
		PyObject *pkey
		PyObject *pval
		Py_ssize_t pos = 0

	rv = factory()
	f = get_map_iter(d, &obj)
	d = <object>obj
	Py_DECREF(d)
	while f(d, &pos, &pkey, &pval):
		rv[func(<object>pkey)] = <object>pval
	return rv

cpdef object itemmap(object func, object d, object factory=dict):
	"""itemmap(func: collections.abc.Callable[[tuple[K, V]], tuple[L, W]], d: collections.abc.Mapping[K, V], factory: collections.abc.Callable[[], collections.abc.MutableMapping[L, W]] = dict) -> collections.abc.MutableMapping[L, W]

	Apply `func` to all items of `d` and return a new `Mapping` with the transformed items.

	(AI generated docstring)

	You can use `itemmap` to transform all keys and values in `d` simultaneously without
	changing `d`. `itemmap` applies `func` to each item yielded by `d.items()`. Each item
	is passed to `func` as a `tuple[K, V]`, and `func` must return a `tuple[L, W]`. `itemmap`
	inserts each returned `tuple` as a new key-value pair in a `MutableMapping`[1] created
	by `factory`.

	Parameters
	----------
	func : Callable[[tuple[K, V]], tuple[L, W]]
		`Callable` applied to each item of `d`. Each item of `d` is passed to `func` as a
		`tuple[K, V]`, and `func` must return a `tuple[L, W]` containing the new key and value.
	d : Mapping[K, V]
		Source `Mapping`[1]. `itemmap` reads all items from `d` and does not change `d`.
	factory : Callable[[], MutableMapping[L, W]] = dict
		`Callable` that creates the `MutableMapping`[1] to `return`.

	Returns
	-------
	mappingTransformed : MutableMapping[L, W]
		New `MutableMapping` created by `factory` populated with each `tuple[L, W]` returned
		by `func`.

	See Also
	--------
	keymap : Apply a `Callable` to all keys of a `Mapping` and return a new `Mapping`.
	valmap : Apply a `Callable` to all values of a `Mapping` and return a new `Mapping`.

	Examples
	--------
	>>> accountids = {'Alice': 10, 'Bob': 20}
	>>> itemmap(reversed, accountids)  # doctest: +SKIP
	{10: 'Alice', 20: 'Bob'}

	References
	----------
	[1] Python `collections.abc` module
		https://docs.python.org/3/library/collections.abc.html
	"""
	cdef:
		object rv, k, v
		f_map_next f
		PyObject *obj
		PyObject *pkey
		PyObject *pval
		Py_ssize_t pos = 0

	rv = factory()
	f = get_map_iter(d, &obj)
	d = <object>obj
	Py_DECREF(d)
	while f(d, &pos, &pkey, &pval):
		k, v = func((<object>pkey, <object>pval))
		rv[k] = v
	return rv

cpdef object valfilter(object predicate, object d, object factory=dict):
	"""valfilter(predicate: collections.abc.Callable[[V], bool], d: collections.abc.Mapping[K, V], factory: collections.abc.Callable[[], collections.abc.MutableMapping[K, V]] = dict) -> collections.abc.MutableMapping[K, V]

	Retain only items from `d` whose values satisfy `predicate` and return a new `Mapping`.

	(AI generated docstring)

	You can use `valfilter` to select items from `d` (***d***ictionary) by their values. `valfilter`
	calls `predicate` with each value yielded by `d.values()`. `valfilter` inserts each item whose
	value causes `predicate` to return `True` into a new `MutableMapping`[1] created by `factory`.
	`valfilter` does not change `d`.

	Parameters
	----------
	predicate : Callable[[V], bool]
		`Callable` applied to each value of `d`. `valfilter` keeps each item for which `predicate`
		returns `True`.
	d : Mapping[K, V]
		Source `Mapping`[1]. `valfilter` reads all items from `d` and does not change `d`.
	factory : Callable[[], MutableMapping[K, V]] = dict
		`Callable` that creates the `MutableMapping`[1] to `return`.

	Returns
	-------
	mappingFiltered : MutableMapping[K, V]
		New `MutableMapping` created by `factory` containing only items from `d` whose values
		cause `predicate` to return `True`.

	See Also
	--------
	keyfilter : Retain only items from `d` whose keys satisfy `predicate` and return a new `Mapping`.
	itemfilter : Retain only items from `d` whose key-value pairs satisfy `predicate` and return a new `Mapping`.
	valmap : Apply a `Callable` to all values of a `Mapping` and return a new `Mapping`.

	>>> iseven = lambda x: x % 2 == 0
	>>> d = {1: 2, 2: 3, 3: 4, 4: 5}
	>>> valfilter(iseven, d)
	{1: 2, 3: 4}

	References
	----------
	[1] Python `collections.abc` module
		https://docs.python.org/3/library/collections.abc.html
	"""
	cdef:
		object rv
		f_map_next f
		PyObject *obj
		PyObject *pkey
		PyObject *pval
		Py_ssize_t pos = 0

	rv = factory()
	f = get_map_iter(d, &obj)
	d = <object>obj
	Py_DECREF(d)
	while f(d, &pos, &pkey, &pval):
		if predicate(<object>pval):
			rv[<object>pkey] = <object>pval
	return rv

cpdef object keyfilter(object predicate, object d, object factory=dict):
	"""keyfilter(predicate: collections.abc.Callable[[K], bool], d: collections.abc.Mapping[K, V], factory: collections.abc.Callable[[], collections.abc.MutableMapping[K, V]] = dict) -> collections.abc.MutableMapping[K, V]

	Retain only items from `d` whose keys satisfy `predicate` and return a new `Mapping`.

	(AI generated docstring)

	You can use `keyfilter` to select items from `d` (***d***ictionary) by their keys. `keyfilter`
	calls `predicate` with each key yielded by `d.keys()`. `keyfilter` inserts each item whose
	key causes `predicate` to return `True` into a new `MutableMapping`[1] created by `factory`.
	`keyfilter` does not change `d`.

	Parameters
	----------
	predicate : Callable[[K], bool]
		`Callable` applied to each key of `d`. `keyfilter` keeps each item for which `predicate`
		returns `True`.
	d : Mapping[K, V]
		Source `Mapping`[1]. `keyfilter` reads all items from `d` and does not change `d`.
	factory : Callable[[], MutableMapping[K, V]] = dict
		`Callable` that creates the `MutableMapping`[1] to `return`.

	Returns
	-------
	mappingFiltered : MutableMapping[K, V]
		New `MutableMapping` created by `factory` containing only items from `d` whose keys
		cause `predicate` to return `True`.

	See Also
	--------
	valfilter : Retain only items from `d` whose values satisfy `predicate` and return a new `Mapping`.
	itemfilter : Retain only items from `d` whose key-value pairs satisfy `predicate` and return a new `Mapping`.
	keymap : Apply a `Callable` to all keys of a `Mapping` and return a new `Mapping`.

	>>> iseven = lambda x: x % 2 == 0
	>>> d = {1: 2, 2: 3, 3: 4, 4: 5}
	>>> keyfilter(iseven, d)
	{2: 3, 4: 5}

	References
	----------
	[1] Python `collections.abc` module
		https://docs.python.org/3/library/collections.abc.html
	"""
	cdef:
		object rv
		f_map_next f
		PyObject *obj
		PyObject *pkey
		PyObject *pval
		Py_ssize_t pos = 0

	rv = factory()
	f = get_map_iter(d, &obj)
	d = <object>obj
	Py_DECREF(d)
	while f(d, &pos, &pkey, &pval):
		if predicate(<object>pkey):
			rv[<object>pkey] = <object>pval
	return rv

cpdef object itemfilter(object predicate, object d, object factory=dict):
	"""itemfilter(predicate: collections.abc.Callable[[tuple[K, V]], bool], d: collections.abc.Mapping[K, V], factory: collections.abc.Callable[[], collections.abc.MutableMapping[K, V]] = dict) -> collections.abc.MutableMapping[K, V]

	Retain only items from `d` whose key-value pairs satisfy `predicate` and return a new `Mapping`.

	(AI generated docstring)

	You can use `itemfilter` to select items from `d` (***d***ictionary) by both key and value
	simultaneously. `itemfilter` calls `predicate` with each item yielded by `d.items()`. Each
	item is passed to `predicate` as a `tuple[K, V]`. `itemfilter` inserts each item for which
	`predicate` returns `True` into a new `MutableMapping`[1] created by `factory`. `itemfilter`
	does not change `d`.

	Parameters
	----------
	predicate : Callable[[tuple[K, V]], bool]
		`Callable` applied to each item of `d`. Each item is passed to `predicate` as a
		`tuple[K, V]`, and `predicate` must return `True` for the item to be retained.
	d : Mapping[K, V]
		Source `Mapping`[1]. `itemfilter` reads all items from `d` and does not change `d`.
	factory : Callable[[], MutableMapping[K, V]] = dict
		`Callable` that creates the `MutableMapping`[1] to `return`.

	Returns
	-------
	mappingFiltered : MutableMapping[K, V]
		New `MutableMapping` created by `factory` containing only items from `d` for which
		`predicate` returns `True`.

	See Also
	--------
	keyfilter : Retain only items from `d` whose keys satisfy `predicate` and return a new `Mapping`.
	valfilter : Retain only items from `d` whose values satisfy `predicate` and return a new `Mapping`.
	itemmap : Apply a `Callable` to all items of a `Mapping` and return a new `Mapping`.

	>>> def isvalid(item):
	...     k, v = item
	...     return k % 2 == 0 and v < 4

	>>> d = {1: 2, 2: 3, 3: 4, 4: 5}
	>>> itemfilter(isvalid, d)
	{2: 3}

	References
	----------
	[1] Python `collections.abc` module
		https://docs.python.org/3/library/collections.abc.html
	"""
	cdef:
		object rv, k, v
		f_map_next f
		PyObject *obj
		PyObject *pkey
		PyObject *pval
		Py_ssize_t pos = 0

	rv = factory()
	f = get_map_iter(d, &obj)
	d = <object>obj
	Py_DECREF(d)
	while f(d, &pos, &pkey, &pval):
		k = <object>pkey
		v = <object>pval
		if predicate((k, v)):
			rv[k] = v
	return rv

cpdef object assoc(object d, object key, object value, object factory=dict):
	"""assoc(d: collections.abc.Mapping[K, V], key: K, value: V, factory: collections.abc.Callable[[], collections.abc.MutableMapping[K, V]] = dict) -> collections.abc.MutableMapping[K, V]

	Create a new `Mapping`[1] with `key` associated with `value`.

	You can use `assoc` (***assoc***iate) to copy `d` (***d***ictionary) to a new `Mapping` created by
	`factory` and assign `value` to `key`. `assoc` does not change `d`.

	Parameters
	----------
	d : Mapping[K, V]
		Source `Mapping`.
	key : K
		`key` that `assoc` inserts or replaces.
	value : V
		`value` that `assoc` assigns to `key`.
	factory : Callable[[], MutableMapping[K, V]] = dict
		`Callable` that creates the `MutableMapping`[1] to `return`.

	Returns
	-------
	mappingUpdated : MutableMapping[K, V]
		New `Mapping` with `key` associated to `value`.

	Examples
	--------
	>>> assoc({}, 'a', 1)
	{'a': 1}
	>>> assoc({'a': 1}, 'a', 3)
	{'a': 3}
	>>> assoc({'a': 1}, 'b', 3)
	{'a': 1, 'b': 3}

	References
	----------
	[1] Python `collections.abc` module
		https://docs.python.org/3/library/collections.abc.html
	"""
	cdef object rv
	rv = factory()
	if PyDict_CheckExact(rv):
		PyDict_Update(rv, d)
	else:
		rv.update(d)
	rv[key] = value
	return rv

cpdef object assoc_in(object d, object keys, object value, object factory=dict):
	"""assoc_in(d: collections.abc.Mapping[K, V], keys: collections.abc.Sequence[K], value: V, factory: collections.abc.Callable[[], collections.abc.MutableMapping[K, V]] = dict) -> collections.abc.MutableMapping[K, V]

	Create a new `MutableMapping` from `d` with `value` at the path specified by `keys`.

	(AI generated docstring)

	You can use `assoc_in` to produce a copy of `d` with `value` placed at the nested
	location given by `keys`. `assoc_in` creates intermediate `MutableMapping` instances
	as needed when a key in `keys` is absent from `d`. `assoc_in` does not mutate `d` or
	any nested `Mapping` within `d`.

	Parameters
	----------
	d : Mapping[K, V]
		Source `Mapping`.
	keys : Iterable[K]
		Non-empty sequence of keys specifying the nested path to the target location in `d`.
	value : V
		The value to place at the location specified by `keys`.
	factory : Callable[[], MutableMapping[K, V]] = dict
		`Callable` that creates each new `MutableMapping`[1] in the result.

	Returns
	-------
	mappingUpdated : MutableMapping[K, V]
		New `MutableMapping` based on `d` with `value` at the path specified by `keys`.

	See Also
	--------
	update_in : Apply a `Callable` to a value at a nested path in a `Mapping`.
	get_in : Retrieve a value at a nested path in a `Mapping`.
	assoc : Create a new `Mapping` from `d` with one key associated to a value.

	Examples
	--------
	>>> assoc_in({'a': 1}, ['a'], 2)
	{'a': 2}
	>>> assoc_in({'a': {'b': 1}}, ['a', 'b'], 2)
	{'a': {'b': 2}}
	>>> assoc_in({}, ['a', 'b'], 1)
	{'a': {'b': 1}}

	References
	----------
	[1] Python `collections.abc` module
		https://docs.python.org/3/library/collections.abc.html
	"""
	cdef object prevkey, key
	cdef object rv, inner, dtemp
	prevkey, keys = keys[0], keys[1:]
	rv = factory()
	if PyDict_CheckExact(rv):
		PyDict_Update(rv, d)
	else:
		rv.update(d)
	inner = rv

	for key in keys:
		if prevkey in d:
			d = d[prevkey]
			dtemp = factory()
			if PyDict_CheckExact(dtemp):
				PyDict_Update(dtemp, d)
			else:
				dtemp.update(d)
		else:
			d = factory()
			dtemp = d
		inner[prevkey] = dtemp
		prevkey = key
		inner = dtemp

	inner[prevkey] = value
	return rv

cdef object c_dissoc(object d, object keys, object factory=dict):
	# implementation copied from humpy_toolz.  Not benchmarked.
	cdef object rv
	rv = factory()
	if len(keys) < len(d) * 0.6:
		rv.update(d)
		for key in keys:
			if key in rv:
				del rv[key]
	else:
		remaining = set(d)
		remaining.difference_update(keys)
		for k in remaining:
			rv[k] = d[k]
	return rv

def dissoc(d, *keys, **kwargs):
	"""dissoc(d: collections.abc.Mapping[K, V], *keys: K, factory: collections.abc.Callable[[], collections.abc.MutableMapping[K, V]] = dict) -> collections.abc.MutableMapping[K, V]

	Create a new `MutableMapping`[1] from `d` with the specified `keys` removed.

	(AI generated docstring)

	You can use `dissoc` (***dissoc***iate) to copy `d` (***d***ictionary) to a new `MutableMapping`
	created by `factory`, then remove each key in `keys` from the result. `dissoc` does not change
	`d`. Keys in `keys` that are absent from `d` are silently ignored.

	Parameters
	----------
	d : Mapping[K, V]
		Source `Mapping`.
	*keys : K
		Keys to remove from `d` in the returned `MutableMapping`.
	factory : Callable[[], MutableMapping[K, V]] = dict
		`Callable` that creates the `MutableMapping`[1] to `return`.

	Returns
	-------
	mappingReduced : MutableMapping[K, V]
		New `MutableMapping` containing all items from `d` except those whose keys are in `keys`.

	Algorithm Details
	-----------------
	`dissoc` selects between two strategies based on the ratio of removed keys to total keys.

	When fewer than 60% of keys are removed, `dissoc` copies all items from `d` and then
	deletes the specified keys one by one.

	When 60% or more of keys are removed, `dissoc` computes the set of remaining keys and
	copies only those items to the result, avoiding unnecessary copy-and-delete operations.

	See Also
	--------
	assoc : Create a new `MutableMapping` from `d` with one key associated to a value.

	Examples
	--------
	>>> dissoc({'x': 1, 'y': 2}, 'y')
	{'x': 1}
	>>> dissoc({'x': 1, 'y': 2}, 'y', 'x')
	{}
	>>> dissoc({'x': 1}, 'y') # Ignores missing keys
	{'x': 1}

	References
	----------
	[1] Python `collections.abc` module
		https://docs.python.org/3/library/collections.abc.html
	"""
	return c_dissoc(d, keys, get_factory('dissoc', kwargs))

cpdef object update_in(object d, object keys, object func, object default=None, object factory=dict):
	"""update_in(d: collections.abc.Mapping[K, V], keys: collections.abc.Sequence[K], func: collections.abc.Callable[[V], V], default: V | None = None, factory: collections.abc.Callable[[], collections.abc.MutableMapping[K, V]] = dict) -> collections.abc.MutableMapping[K, V]

	Apply a `Callable` to a value at a nested path in a `Mapping`.

	(AI generated docstring)

	You can use `update_in` to produce a copy of `d` with the value at the nested path
	specified by `keys` replaced by the result of calling `func` on the current value.
	`update_in` creates intermediate `MutableMapping` instances as needed when a key in
	`keys` is absent from `d`. If the innermost key is absent, `func` receives `default`
	instead of an existing value. `update_in` does not mutate `d` or any nested `Mapping`
	within `d`.

	Parameters
	----------
	d : Mapping[K, V]
		Source `Mapping`.
	keys : Sequence[K]
		Non-empty sequence of keys specifying the nested path to the value to update in `d`.
	func : Callable[[V], V]
		`Callable` applied to the current value at the path in `keys`. If the innermost
		key is absent from `d`, `func` receives `default`.
	default : V | None = None
		Value passed to `func` when the innermost key is absent from `d`.
	factory : Callable[[], MutableMapping[K, V]] = dict
		`Callable` that creates each new `MutableMapping`[1] in the result.

	Returns
	-------
	mappingUpdated : MutableMapping[K, V]
		New `MutableMapping` based on `d` with the value at the path specified by `keys`
		replaced by the result of `func`.

	See Also
	--------
	assoc_in : Create a new `Mapping` from `d` with a value at a nested path.
	get_in : Retrieve a value at a nested path in a `Mapping`.
	assoc : Create a new `Mapping` from `d` with one key associated to a value.

	Examples
	--------
	>>> inc = lambda x: x + 1
	>>> update_in({'a': 0}, ['a'], inc)
	{'a': 1}

	>>> update_in({}, ['z'], inc, 0)
	{'z': 1}

	>>> update_in({}, [1, 2, 3], str, default="bar")
	{1: {2: {3: 'bar'}}}

	References
	----------
	[1] Python `collections.abc` module
		https://docs.python.org/3/library/collections.abc.html
	"""
	cdef object prevkey, key
	cdef object rv, inner, dtemp
	prevkey, keys = keys[0], keys[1:]
	rv = factory()
	if PyDict_CheckExact(rv):
		PyDict_Update(rv, d)
	else:
		rv.update(d)
	inner = rv

	for key in keys:
		if prevkey in d:
			d = d[prevkey]
			dtemp = factory()
			if PyDict_CheckExact(dtemp):
				PyDict_Update(dtemp, d)
			else:
				dtemp.update(d)
		else:
			d = factory()
			dtemp = d
		inner[prevkey] = dtemp
		prevkey = key
		inner = dtemp

	if prevkey in d:
		key = func(d[prevkey])
	else:
		key = func(default)
	inner[prevkey] = key
	return rv

cdef tuple _get_in_exceptions = (KeyError, IndexError, TypeError)

cpdef object get_in(object keys, object coll, object default=None, object no_default=False):
	"""get_in(keys: collections.abc.Sequence[K], coll: humpy_toolz.dicttoolz.SupportsGetItem[K, V], default: Optional[V] = None, no_default: bool = False) -> Optional[V]

	Retrieve a value from a potentially nested `coll` (***coll***ection) using a `Sequence` of `keys`.

	(AI generated docstring)

	You can use `get_in` to navigate into a nested `coll` (***coll***ection) by following a
	sequence of `keys`. `get_in` applies each key in `keys` sequentially using
	`operator.getitem`[1]. If the path does not exist, `get_in` returns `default`. If
	`no_default` is `True`, `get_in` re-raises the original exception instead of returning
	`default`.

	Parameters
	----------
	keys : Sequence[K]
		Sequence of keys that describes the path to traverse in `coll`.
	coll : SupportsGetItem[K, V]
		Collection to traverse. `get_in` applies each key in `keys` to the current `coll`
		using `operator.getitem`[1], so `coll` can be any nested structure such as a `dict`
		or `list`.
	default : V | None = None
		Value to return when the path in `keys` does not exist in `coll`.
	no_default : bool = False
		When `True`, re-raise the original `KeyError`, `IndexError`, or `TypeError`
		instead of returning `default`.

	Returns
	-------
	value : V | None
		The value at the nested path in `coll`, or `default` if the path does not exist.

	Raises
	------
	KeyError
		When `no_default` is `True` and a key is missing from a mapping.
	IndexError
		When `no_default` is `True` and an index is out of range.
	TypeError
		When `no_default` is `True` and a key type is incompatible with `coll`.

	See Also
	--------
	itertoolz.get : Retrieve a value or values from a collection.
	operator.getitem : Return the value at a given key in a collection.
	assoc_in : Create a new `Mapping` from `d` with a value at a nested path.
	update_in : Apply a `Callable` to a value at a nested path in a `Mapping`.

	Examples
	--------
	>>> transaction = {'name': 'Alice',
	...                'purchase': {'items': ['Apple', 'Orange'],
	...                             'costs': [0.50, 1.25]},
	...                'credit card': '5555-1234-1234-1234'}
	>>> get_in(['purchase', 'items', 0], transaction)
	'Apple'
	>>> get_in(['name'], transaction)
	'Alice'
	>>> get_in(['purchase', 'total'], transaction)
	>>> get_in(['purchase', 'items', 'apple'], transaction)
	>>> get_in(['purchase', 'items', 10], transaction)
	>>> get_in(['purchase', 'total'], transaction, 0)
	0
	>>> get_in(['y'], {}, no_default=True)
	Traceback (most recent call last):
		...
	KeyError: 'y'

	References
	----------
	[1] Python `operator` module
		https://docs.python.org/3/library/operator.html#operator.getitem
	"""
	cdef object item
	try:
		for item in keys:
			coll = coll[item]
		return coll
	except _get_in_exceptions:
		if no_default:
			raise
		return default



from collections.abc import Callable, Mapping, MutableMapping
from typing import overload

__all__ = ["merge", "merge_with"]

@overload
def merge_with[K, V]() -> Callable[
	..., dict[K, V] | MutableMapping[K, V]
]: ...
@overload
def merge_with[K, V](
	func: Callable[[list[V]], V], /
) -> Callable[..., dict[K, V] | MutableMapping[K, V]]: ...
@overload
def merge_with[K, V](
	func: Callable[[list[V]], V],
	d: Mapping[K, V],
	/,
) -> dict[K, V]: ...
@overload
def merge_with[K, V](
	func: Callable[[list[V]], V],
	d: Mapping[K, V],
	d2: Mapping[K, V],
	/,
	*dicts: Mapping[K, V],
) -> dict[K, V]: ...
@overload
def merge_with[K, V](
	func: Callable[[list[V]], V],
	d: Mapping[K, V],
	d2: Mapping[K, V],
	/,
	*dicts: Mapping[K, V],
	factory: Callable[[], MutableMapping[K, V]],
) -> MutableMapping[K, V]: ...
@overload
def merge_with[K, V](
	func: Callable[[list[V]], V],
	/,
	*,
	factory: Callable[[], MutableMapping[K, V]],
) -> Callable[..., MutableMapping[K, V]]: ...
def merge_with[K, V](
	func: Callable[[list[V]], V] = ...,
	d: Mapping[K, V] = ...,
	*dicts: Mapping[K, V],
	factory: Callable[[], MutableMapping[K, V]] = ...,
) -> (
	dict[K, V]
	| MutableMapping[K, V]
	| Callable[..., dict[K, V] | MutableMapping[K, V]]
):
	...

@overload
def merge[K, V]() -> Callable[
	..., dict[K, V] | MutableMapping[K, V]
]: ...
@overload
def merge[K, V](d: Mapping[K, V], /) -> dict[K, V]: ...
@overload
def merge[K, V](
	d: Mapping[K, V],
	d2: Mapping[K, V],
	/,
	*dicts: Mapping[K, V],
) -> dict[K, V]: ...
@overload
def merge[K, V](
	d: Mapping[K, V],
	d2: Mapping[K, V],
	/,
	*dicts: Mapping[K, V],
	factory: Callable[[], MutableMapping[K, V]],
) -> MutableMapping[K, V]: ...
@overload
def merge[K, V](
	*,
	factory: Callable[[], MutableMapping[K, V]],
) -> Callable[..., MutableMapping[K, V]]: ...
def merge[K, V](
	d: Mapping[K, V] = ...,
	*dicts: Mapping[K, V],
	factory: Callable[[], MutableMapping[K, V]] = ...,
) -> (
	dict[K, V]
	| MutableMapping[K, V]
	| Callable[..., dict[K, V] | MutableMapping[K, V]]
):
	...



from collections.abc import Callable, Iterable
from typing import Literal

type _MapFunction[T] = Callable[
	[
		Callable[[Iterable[T]], T],
		Iterable[Iterable[T]],
	],
	Iterable[T],
]

def fold[T](
	binop: Callable[[T, T], T],
	seq: Iterable[T],
	default: Literal[__no__default__] | T = "__no_default__",
	map: _MapFunction[T] = map,
	chunksize: int = 128,
	combine: Callable[[T, T], T] | None = None,
) -> T:
	...



cdef class remove:
    cdef object predicate
    cdef object iter_seq


cdef class accumulate:
    cdef object binop
    cdef object iter_seq
    cdef object result
    cdef object initial


cpdef dict groupby(object key, object seq)


cdef class _merge_sorted:
    cdef object seq1
    cdef object seq2
    cdef object val1
    cdef object val2
    cdef Py_ssize_t loop

cdef class _merge_sorted_key:
    cdef object seq1
    cdef object seq2
    cdef object val1
    cdef object val2
    cdef object key
    cdef object key1
    cdef object key2
    cdef Py_ssize_t loop


cdef object c_merge_sorted(object seqs, object key=*)


cdef class interleave:
    cdef list iters
    cdef list newiters
    cdef Py_ssize_t i
    cdef Py_ssize_t n


cdef class _unique_key:
    cdef object key
    cdef object iter_seq
    cdef object seen


cdef class _unique_identity:
    cdef object iter_seq
    cdef object seen


cpdef object unique(object seq, object key=*)


cpdef object isiterable(object x)


cpdef object isdistinct(object seq)


cpdef object take(Py_ssize_t n, object seq)


cpdef object tail(Py_ssize_t n, object seq)


cpdef object drop(Py_ssize_t n, object seq)


cpdef object take_nth(Py_ssize_t n, object seq)


cpdef object first(object seq)


cpdef object second(object seq)


cpdef object nth(Py_ssize_t n, object seq)


cpdef object last(object seq)


cpdef object rest(object seq)


cpdef object get(object ind, object seq, object default=*)


cpdef object cons(object el, object seq)


cpdef object concat(object seqs)


cpdef object mapcat(object func, object seqs)


cdef class interpose:
    cdef object el
    cdef object iter_seq
    cdef object val
    cdef bint do_el


cpdef dict frequencies(object seq)


cpdef dict reduceby(object key, object binop, object seq, object init=*)


cdef class iterate:
    cdef object func
    cdef object x
    cdef object val


cdef class sliding_window:
    cdef object iterseq
    cdef tuple prev
    cdef Py_ssize_t n


cpdef object partition(Py_ssize_t n, object seq, object pad=*)


cdef class partition_all:
    cdef Py_ssize_t n
    cdef object iterseq
    cdef object seq


cpdef object count(object seq)


cdef class _pluck_index:
    cdef object ind
    cdef object iterseqs


cdef class _pluck_index_default:
    cdef object ind
    cdef object iterseqs
    cdef object default


cdef class _pluck_list:
    cdef list ind
    cdef object iterseqs
    cdef Py_ssize_t n


cdef class _pluck_list_default:
    cdef list ind
    cdef object iterseqs
    cdef object default
    cdef Py_ssize_t n


cpdef object pluck(object ind, object seqs, object default=*)


cdef class _getter_index:
    cdef object ind


cdef class _getter_list:
    cdef list ind
    cdef Py_ssize_t n


cdef class _getter_null:
    pass


cpdef object getter(object index)


cpdef object join(object leftkey, object leftseq,
                  object rightkey, object rightseq,
                  object left_default=*,
                  object right_default=*)

cdef class _join:
    cdef dict d
    cdef list matches
    cdef set seen_keys
    cdef object leftseq
    cdef object rightseq
    cdef object _rightkey
    cdef object right
    cdef object left_default
    cdef object right_default
    cdef object keys
    cdef Py_ssize_t N
    cdef Py_ssize_t i
    cdef bint is_rightseq_exhausted

    cdef object rightkey(self)


cdef class _inner_join(_join):
    pass

cdef class _right_outer_join(_join):
    pass

cdef class _left_outer_join(_join):
    pass

cdef class _outer_join(_join):
    pass


cdef class _inner_join_key(_inner_join):
    pass

cdef class _inner_join_index(_inner_join):
    pass

cdef class _inner_join_indices(_inner_join):
    pass

cdef class _right_outer_join_key(_right_outer_join):
    pass

cdef class _right_outer_join_index(_right_outer_join):
    pass

cdef class _right_outer_join_indices(_right_outer_join):
    pass

cdef class _left_outer_join_key(_left_outer_join):
    pass

cdef class _left_outer_join_index(_left_outer_join):
    pass

cdef class _left_outer_join_indices(_left_outer_join):
    pass

cdef class _outer_join_key(_outer_join):
    pass

cdef class _outer_join_index(_outer_join):
    pass

cdef class _outer_join_indices(_outer_join):
    pass


cdef class _diff_key:
    cdef Py_ssize_t N
    cdef object iters
    cdef object key


cdef class _diff_identity:
    cdef Py_ssize_t N
    cdef object iters


cdef object c_diff(object seqs, object default=*, object key=*)


cpdef object topk(Py_ssize_t k, object seq, object key=*)


cpdef object peek(object seq)


cpdef object peekn(Py_ssize_t n, object seq)


cdef class random_sample:
    cdef object iter_seq
    cdef object prob
    cdef object random_func



from collections import defaultdict as _defaultdict
from collections.abc import Callable, ItemsView, Iterator, KeysView, Mapping, MutableMapping, ValuesView
from humpy_toolz.dicttoolz import (
	assoc, assoc_in, dissoc, get_in, itemfilter, itemmap, keyfilter, keymap, merge, merge_with, update_in, valfilter, valmap)
from humpy_toolz.functoolz import identity
from humpy_toolz.utils import raises
from typing import Any, ClassVar
import os
import pytest

def inc(x: int) -> int:
	return x + 1

def iseven(i: int) -> bool:
	return i % 2 == 0

def bucketKeyByParity(key: int) -> int:
	return key % 2

def swapItemKeyWithValue(item: tuple[int, int]) -> tuple[int, int]:
	return item[1], item[0]

def summarizeItemToParityAndTotal(item: tuple[int, int]) -> tuple[int, int]:
	return item[0] % 2, item[0] + item[1]

def valueExceedsThreeHundred(value: int) -> bool:
	return value > 300

def keyExceedsThreeHundred(key: int) -> bool:
	return key > 300

def itemHasEvenKeyAndOddValue(item: tuple[int, int]) -> bool:
	return item[0] % 2 == 0 and item[1] % 2 == 1

def itemHasLargeKeyAndLargeValue(item: tuple[int, int]) -> bool:
	return item[0] > 300 and item[1] > 300

def makeDefaultDictFactory(
	itemIterable: ItemsView[Any, Any] | Iterator[tuple[Any, Any]] | None = None,
) -> 'defaultdict[Any, Any]':
	if itemIterable is None:
		return defaultdict(int)
	return defaultdict(int, itemIterable)

def makeCustomMappingFactory(
	itemIterable: ItemsView[Any, Any] | Iterator[tuple[Any, Any]] | None = None,
) -> 'CustomMapping':
	if itemIterable is None:
		return CustomMapping()
	return CustomMapping(itemIterable)

class TestDict:
	"""Test typical usage: dict inputs, no factory keyword.

	Class attributes:
		D: callable that inputs a dict and creates or returns a MutableMapping
		kw: kwargs dict to specify "factory" keyword (if applicable)
	"""
	D: ClassVar[Callable[..., Any]] = dict
	kw: ClassVar[dict[str, Any]] = {}

	def test_merge(self) -> None:
		D, kw = (self.D, self.kw)
		assert merge(D({1: 1, 2: 2}), D({3: 4}), **kw) == D({1: 1, 2: 2, 3: 4})

	def test_merge_iterable_arg(self) -> None:
		D, kw = (self.D, self.kw)
		assert merge([D({1: 1, 2: 2}), D({3: 4})], **kw) == D({1: 1, 2: 2, 3: 4})  # pyright: ignore[reportCallIssue, reportArgumentType]  # ty:ignore[no-matching-overload]

	def test_merge_three_dicts(self) -> None:
		D, kw = (self.D, self.kw)
		assert merge(D({2: 3}), D({5: 7}), D({11: 13}), **kw) == D({2: 3, 5: 7, 11: 13})
		assert merge(D({2: 3, 5: 7}), D({5: 11, 11: 13}), D({2: 17, 11: 19}), **kw) == D({2: 17, 5: 11, 11: 19})

	def test_merge_with(self) -> None:
		D, kw = (self.D, self.kw)
		dicts = (D({1: 1, 2: 2}), D({1: 10, 2: 20}))
		assert merge_with(sum, *dicts, **kw) == D({1: 11, 2: 22})
		assert merge_with(tuple, *dicts, **kw) == D({1: (1, 10), 2: (2, 20)})
		dicts = (D({1: 1, 2: 2, 3: 3}), D({1: 10, 2: 20}))
		assert merge_with(sum, *dicts, **kw) == D({1: 11, 2: 22, 3: 3})
		assert merge_with(tuple, *dicts, **kw) == D({1: (1, 10), 2: (2, 20), 3: (3,)})
		assert not merge_with(sum)  # pyright: ignore[reportArgumentType]

	def test_merge_with_iterable_arg(self) -> None:
		D, kw = (self.D, self.kw)
		dicts = (D({1: 1, 2: 2}), D({1: 10, 2: 20}))
		assert merge_with(sum, *dicts, **kw) == D({1: 11, 2: 22})
		assert merge_with(sum, dicts, **kw) == D({1: 11, 2: 22})  # pyright: ignore[reportCallIssue, reportArgumentType]  # ty:ignore[no-matching-overload]
		assert merge_with(sum, iter(dicts), **kw) == D({1: 11, 2: 22})  # pyright: ignore[reportCallIssue, reportArgumentType]  # ty:ignore[no-matching-overload]

	def test_merge_with_three_dicts(self) -> None:
		D, kw = (self.D, self.kw)
		dicts = (D({2: 3, 5: 7}), D({2: 11, 5: 13}), D({2: 17}))
		assert merge_with(sum, *dicts, **kw) == D({2: 31, 5: 20})
		assert merge_with(tuple, *dicts, **kw) == D({2: (3, 11, 17), 5: (7, 13)})

	@pytest.mark.parametrize(
		('functionUnderTest', 'sourceMappingDefinition', 'expectedMappingDefinition'),
		(
			pytest.param(
				inc,
				{13: 21, 34: 55},
				{13: 22, 34: 56},
				id='increments each value',
			),
			pytest.param(
				str,
				{13: 21, 34: 55},
				{13: '21', 34: '55'},
				id='converts each value to text',
			),
		),
	)
	def test_valmap(
		self,
		functionUnderTest: Callable[[Any], Any],
		sourceMappingDefinition: dict[int, Any],
		expectedMappingDefinition: dict[int, Any],
	) -> None:
		mappingFactory: Callable[[dict[Any, Any]], Any] = self.D
		factoryKeywordArguments: dict[str, Any] = self.kw
		sourceMapping: Any = mappingFactory(sourceMappingDefinition)
		expectedMapping: Any = mappingFactory(expectedMappingDefinition)

		resultMapping: Any = valmap(functionUnderTest, sourceMapping, **factoryKeywordArguments)

		assert resultMapping == expectedMapping, (
			f'valmap returned {resultMapping}, expected {expectedMapping} for source '
			f'{sourceMappingDefinition}.'
		)
		assert sourceMapping == mappingFactory(sourceMappingDefinition), (
			f'valmap mutated source mapping {sourceMapping}, expected {sourceMappingDefinition}.'
		)
		assert resultMapping is not sourceMapping, (
			f'valmap returned source mapping {sourceMapping} instead of a new mapping for '
			f'{sourceMappingDefinition}.'
		)
		assert isinstance(resultMapping, type(expectedMapping)), (
			f'valmap returned {type(resultMapping).__name__}, expected '
			f'{type(expectedMapping).__name__} for source {sourceMappingDefinition}.'
		)

	@pytest.mark.parametrize(
		('functionUnderTest', 'sourceMappingDefinition', 'expectedMappingDefinition'),
		(
			pytest.param(
				inc,
				{13: 21, 34: 55},
				{14: 21, 35: 55},
				id='increments each key',
			),
			pytest.param(
				bucketKeyByParity,
				{13: 21, 34: 55, 55: 89},
				{1: 89, 0: 55},
				id='keeps last value when transformed keys collide',
			),
		),
	)
	def test_keymap(
		self,
		functionUnderTest: Callable[[Any], Any],
		sourceMappingDefinition: dict[int, Any],
		expectedMappingDefinition: dict[int, Any],
	) -> None:
		mappingFactory: Callable[[dict[Any, Any]], Any] = self.D
		factoryKeywordArguments: dict[str, Any] = self.kw
		sourceMapping: Any = mappingFactory(sourceMappingDefinition)
		expectedMapping: Any = mappingFactory(expectedMappingDefinition)

		resultMapping: Any = keymap(functionUnderTest, sourceMapping, **factoryKeywordArguments)

		assert resultMapping == expectedMapping, (
			f'keymap returned {resultMapping}, expected {expectedMapping} for source '
			f'{sourceMappingDefinition}.'
		)
		assert sourceMapping == mappingFactory(sourceMappingDefinition), (
			f'keymap mutated source mapping {sourceMapping}, expected {sourceMappingDefinition}.'
		)
		assert resultMapping is not sourceMapping, (
			f'keymap returned source mapping {sourceMapping} instead of a new mapping for '
			f'{sourceMappingDefinition}.'
		)
		assert isinstance(resultMapping, type(expectedMapping)), (
			f'keymap returned {type(resultMapping).__name__}, expected '
			f'{type(expectedMapping).__name__} for source {sourceMappingDefinition}.'
		)

	@pytest.mark.parametrize(
		('functionUnderTest', 'sourceMappingDefinition', 'expectedMappingDefinition'),
		(
			pytest.param(
				swapItemKeyWithValue,
				{13: 21, 34: 55},
				{21: 13, 55: 34},
				id='swaps keys with values',
			),
			pytest.param(
				summarizeItemToParityAndTotal,
				{13: 21, 34: 55, 55: 89},
				{1: 144, 0: 89},
				id='keeps last transformed item when keys collide',
			),
		),
	)
	def test_itemmap(
		self,
		functionUnderTest: Callable[[tuple[Any, Any]], tuple[Any, Any]],
		sourceMappingDefinition: dict[int, Any],
		expectedMappingDefinition: dict[int, Any],
	) -> None:
		mappingFactory: Callable[[dict[Any, Any]], Any] = self.D
		factoryKeywordArguments: dict[str, Any] = self.kw
		sourceMapping: Any = mappingFactory(sourceMappingDefinition)
		expectedMapping: Any = mappingFactory(expectedMappingDefinition)

		resultMapping: Any = itemmap(functionUnderTest, sourceMapping, **factoryKeywordArguments)

		assert resultMapping == expectedMapping, (
			f'itemmap returned {resultMapping}, expected {expectedMapping} for source '
			f'{sourceMappingDefinition}.'
		)
		assert sourceMapping == mappingFactory(sourceMappingDefinition), (
			f'itemmap mutated source mapping {sourceMapping}, expected {sourceMappingDefinition}.'
		)
		assert resultMapping is not sourceMapping, (
			f'itemmap returned source mapping {sourceMapping} instead of a new mapping for '
			f'{sourceMappingDefinition}.'
		)
		assert isinstance(resultMapping, type(expectedMapping)), (
			f'itemmap returned {type(resultMapping).__name__}, expected '
			f'{type(expectedMapping).__name__} for source {sourceMappingDefinition}.'
		)

	@pytest.mark.parametrize(
		('predicateUnderTest', 'sourceMappingDefinition', 'expectedMappingDefinition'),
		(
			pytest.param(
				iseven,
				{13: 21, 34: 144, 55: 233},
				{34: 144},
				id='keeps matching values',
			),
			pytest.param(
				valueExceedsThreeHundred,
				{13: 21, 34: 144, 55: 233},
				{},
				id='returns empty mapping when nothing matches',
			),
		),
	)
	def test_valfilter(
		self,
		predicateUnderTest: Callable[[Any], bool],
		sourceMappingDefinition: dict[int, Any],
		expectedMappingDefinition: dict[int, Any],
	) -> None:
		mappingFactory: Callable[[dict[Any, Any]], Any] = self.D
		factoryKeywordArguments: dict[str, Any] = self.kw
		sourceMapping: Any = mappingFactory(sourceMappingDefinition)
		expectedMapping: Any = mappingFactory(expectedMappingDefinition)

		resultMapping: Any = valfilter(predicateUnderTest, sourceMapping, **factoryKeywordArguments)

		assert resultMapping == expectedMapping, (
			f'valfilter returned {resultMapping}, expected {expectedMapping} for source '
			f'{sourceMappingDefinition}.'
		)
		assert sourceMapping == mappingFactory(sourceMappingDefinition), (
			f'valfilter mutated source mapping {sourceMapping}, expected {sourceMappingDefinition}.'
		)
		assert resultMapping is not sourceMapping, (
			f'valfilter returned source mapping {sourceMapping} instead of a new mapping for '
			f'{sourceMappingDefinition}.'
		)
		assert isinstance(resultMapping, type(expectedMapping)), (
			f'valfilter returned {type(resultMapping).__name__}, expected '
			f'{type(expectedMapping).__name__} for source {sourceMappingDefinition}.'
		)

	@pytest.mark.parametrize(
		('predicateUnderTest', 'sourceMappingDefinition', 'expectedMappingDefinition'),
		(
			pytest.param(
				iseven,
				{13: 21, 34: 55, 55: 89, 144: 233},
				{34: 55, 144: 233},
				id='keeps matching keys',
			),
			pytest.param(
				keyExceedsThreeHundred,
				{13: 21, 34: 55, 55: 89},
				{},
				id='returns empty mapping when no keys match',
			),
		),
	)
	def test_keyfilter(
		self,
		predicateUnderTest: Callable[[Any], bool],
		sourceMappingDefinition: dict[int, Any],
		expectedMappingDefinition: dict[int, Any],
	) -> None:
		mappingFactory: Callable[[dict[Any, Any]], Any] = self.D
		factoryKeywordArguments: dict[str, Any] = self.kw
		sourceMapping: Any = mappingFactory(sourceMappingDefinition)
		expectedMapping: Any = mappingFactory(expectedMappingDefinition)

		resultMapping: Any = keyfilter(predicateUnderTest, sourceMapping, **factoryKeywordArguments)

		assert resultMapping == expectedMapping, (
			f'keyfilter returned {resultMapping}, expected {expectedMapping} for source '
			f'{sourceMappingDefinition}.'
		)
		assert sourceMapping == mappingFactory(sourceMappingDefinition), (
			f'keyfilter mutated source mapping {sourceMapping}, expected {sourceMappingDefinition}.'
		)
		assert resultMapping is not sourceMapping, (
			f'keyfilter returned source mapping {sourceMapping} instead of a new mapping for '
			f'{sourceMappingDefinition}.'
		)
		assert isinstance(resultMapping, type(expectedMapping)), (
			f'keyfilter returned {type(resultMapping).__name__}, expected '
			f'{type(expectedMapping).__name__} for source {sourceMappingDefinition}.'
		)

	@pytest.mark.parametrize(
		('predicateUnderTest', 'sourceMappingDefinition', 'expectedMappingDefinition'),
		(
			pytest.param(
				itemHasEvenKeyAndOddValue,
				{13: 21, 34: 55, 55: 144, 144: 233},
				{34: 55, 144: 233},
				id='filters on both key and value',
			),
			pytest.param(
				itemHasLargeKeyAndLargeValue,
				{13: 21, 34: 55, 55: 144},
				{},
				id='returns empty mapping when no items match',
			),
		),
	)
	def test_itemfilter(
		self,
		predicateUnderTest: Callable[[tuple[Any, Any]], bool],
		sourceMappingDefinition: dict[int, Any],
		expectedMappingDefinition: dict[int, Any],
	) -> None:
		mappingFactory: Callable[[dict[Any, Any]], Any] = self.D
		factoryKeywordArguments: dict[str, Any] = self.kw
		sourceMapping: Any = mappingFactory(sourceMappingDefinition)
		expectedMapping: Any = mappingFactory(expectedMappingDefinition)

		resultMapping: Any = itemfilter(predicateUnderTest, sourceMapping, **factoryKeywordArguments)

		assert resultMapping == expectedMapping, (
			f'itemfilter returned {resultMapping}, expected {expectedMapping} for source '
			f'{sourceMappingDefinition}.'
		)
		assert sourceMapping == mappingFactory(sourceMappingDefinition), (
			f'itemfilter mutated source mapping {sourceMapping}, expected {sourceMappingDefinition}.'
		)
		assert resultMapping is not sourceMapping, (
			f'itemfilter returned source mapping {sourceMapping} instead of a new mapping for '
			f'{sourceMappingDefinition}.'
		)
		assert isinstance(resultMapping, type(expectedMapping)), (
			f'itemfilter returned {type(resultMapping).__name__}, expected '
			f'{type(expectedMapping).__name__} for source {sourceMappingDefinition}.'
		)

	@pytest.mark.parametrize(
		('sourceMappingDefinition', 'keyToAssociate', 'valueToAssociate', 'expectedMappingDefinition'),
		(
			pytest.param(
				{},
				'alpha',
				13,
				{'alpha': 13},
				id='associates into empty mapping',
			),
			pytest.param(
				{'alpha': 13, 'beta': 21},
				'beta',
				34,
				{'alpha': 13, 'beta': 34},
				id='replaces existing value',
			),
			pytest.param(
				{'alpha': 13, 'beta': 21},
				'gamma',
				55,
				{'alpha': 13, 'beta': 21, 'gamma': 55},
				id='preserves siblings when inserting',
			),
		),
	)
	def test_assoc(
		self,
		sourceMappingDefinition: dict[str, int],
		keyToAssociate: str,
		valueToAssociate: int,
		expectedMappingDefinition: dict[str, int],
	) -> None:
		mappingFactory: Callable[[dict[Any, Any]], Any] = self.D
		factoryKeywordArguments: dict[str, Any] = self.kw
		sourceMapping: Any = mappingFactory(sourceMappingDefinition)
		expectedMapping: Any = mappingFactory(expectedMappingDefinition)

		resultMapping: Any = assoc(sourceMapping, keyToAssociate, valueToAssociate, **factoryKeywordArguments)

		assert resultMapping == expectedMapping, (
			f'assoc returned {resultMapping}, expected {expectedMapping} for source '
			f'{sourceMappingDefinition}, key {keyToAssociate}, and value {valueToAssociate}.'
		)
		assert sourceMapping == mappingFactory(sourceMappingDefinition), (
			f'assoc mutated source mapping {sourceMapping}, expected {sourceMappingDefinition}.'
		)
		assert resultMapping is not sourceMapping, (
			f'assoc returned source mapping {sourceMapping} instead of a new mapping for '
			f'{sourceMappingDefinition}.'
		)
		assert isinstance(resultMapping, type(expectedMapping)), (
			f'assoc returned {type(resultMapping).__name__}, expected '
			f'{type(expectedMapping).__name__} for source {sourceMappingDefinition}.'
		)

	@pytest.mark.parametrize(
		('sourceMappingDefinition', 'keysToRemove', 'expectedMappingDefinition'),
		(
			pytest.param(
				{'north': 13, 'south': 21, 'east': 34},
				('south',),
				{'north': 13, 'east': 34},
				id='removes single key',
			),
			pytest.param(
				{'north': 13, 'south': 21, 'east': 34, 'west': 55},
				('north', 'west'),
				{'south': 21, 'east': 34},
				id='removes multiple keys',
			),
			pytest.param(
				{'alpha': 13, 'beta': 21, 'gamma': 34},
				('beta', 'omega'),
				{'alpha': 13, 'gamma': 34},
				id='ignores missing keys',
			),
			pytest.param(
				{'alpha': 13, 'beta': 21, 'gamma': 34, 'delta': 55, 'epsilon': 89},
				('alpha', 'beta', 'gamma'),
				{'delta': 55, 'epsilon': 89},
				id='removes sixty percent boundary',
			),
			pytest.param(
				{'alpha': 13, 'beta': 21, 'gamma': 34},
				(),
				{'alpha': 13, 'beta': 21, 'gamma': 34},
				id='keeps mapping when no keys provided',
			),
			pytest.param(
				{'north': 13, 'south': 21, 'east': 34},
				('south', 'north', 'south', 'east'),
				{},
				id='removes all keys even with duplicates',
			),
		),
	)
	def test_dissocRemovesKeys(
		self,
		sourceMappingDefinition: dict[str, int],
		keysToRemove: tuple[str, ...],
		expectedMappingDefinition: dict[str, int],
	) -> None:
		mappingFactory = self.D
		factoryKeywordArguments: dict[str, Callable[[], MutableMapping[str, int]]] = self.kw
		sourceMapping = mappingFactory(sourceMappingDefinition)
		expectedMapping = mappingFactory(expectedMappingDefinition)

		resultMapping = dissoc(sourceMapping, *keysToRemove, **factoryKeywordArguments)

		assert resultMapping == expectedMapping, (
			f'dissoc returned {resultMapping}, expected {expectedMapping} for keys {keysToRemove} '
			f'and source {sourceMappingDefinition}.'
		)
		assert sourceMapping == mappingFactory(sourceMappingDefinition), (
			f'dissoc mutated source mapping {sourceMapping} for keys {keysToRemove}, '
			f'expected {sourceMappingDefinition}.'
		)
		assert resultMapping is not sourceMapping, (
			f'dissoc returned {resultMapping} which is the same object as {sourceMapping} '
			f'for keys {keysToRemove} and source {sourceMappingDefinition}, expected a new mapping.'
		)
		factoryCallable: Callable[[], MutableMapping[str, int]]
		if 'factory' in factoryKeywordArguments:
			factoryCallable = factoryKeywordArguments['factory']
		else:

			def defaultFactory() -> dict[str, int]:
				return {}

			factoryCallable = defaultFactory

		expectedMappingType = type(factoryCallable())
		assert isinstance(resultMapping, expectedMappingType), (
			f'dissoc returned {type(resultMapping).__name__}, expected {expectedMappingType.__name__} '
			f'for keys {keysToRemove} and source {sourceMappingDefinition}.'
		)

		repeatedResultMapping = dissoc(resultMapping, *keysToRemove, **factoryKeywordArguments)
		assert repeatedResultMapping == resultMapping, (
			f'dissoc was not idempotent for keys {keysToRemove} and source '
			f'{sourceMappingDefinition}, got {repeatedResultMapping}.'
		)

	def test_assoc_in(self) -> None:
		D, kw = (self.D, self.kw)
		assert assoc_in(D({'a': 1}), ['a'], 2, **kw) == D({'a': 2})
		assert assoc_in(D({'a': D({'b': 1})}), ['a', 'b'], 2, **kw) == D({'a': D({'b': 2})})
		assert assoc_in(D({}), ['a', 'b'], 1, **kw) == D({'a': D({'b': 1})})
		d = D({'x': 1})
		oldd = d
		d2 = assoc_in(d, ['x'], 2, **kw)
		assert d is oldd
		assert d2 is not oldd
		# Three-level nesting and new key creation
		assert assoc_in(D({}), ['a', 'b', 'c'], 42, **kw) == D({'a': D({'b': D({'c': 42})})})
		# Existing nested mapping is not mutated
		inner = D({'b': 1})
		d3 = D({'a': inner})
		d4 = assoc_in(d3, ['a', 'b'], 99, **kw)
		assert inner['b'] == 1
		assert d4['a']['b'] == 99
		# Sibling keys at outer level are preserved
		assert assoc_in(D({'a': 1, 'b': 2}), ['a'], 99, **kw) == D({'a': 99, 'b': 2})

	def test_get_in(self) -> None:
		transaction: Any = {
			'name': 'Alice',
			'purchase': {'items': ['Apple', 'Orange'], 'costs': [0.50, 1.25]},
			'credit card': '5555-1234-1234-1234',
		}
		assert get_in(['name'], transaction) == 'Alice'
		assert get_in(['purchase', 'items', 0], transaction) == 'Apple'  # ty:ignore[invalid-argument-type]
		assert get_in(['purchase', 'total'], transaction) is None
		assert get_in(['purchase', 'total'], transaction, 0) == 0
		assert get_in(['purchase', 'items', 'apple'], transaction) is None
		assert get_in(['purchase', 'items', 10], transaction) is None  # ty:ignore[invalid-argument-type]
		assert get_in([], transaction) is transaction
		assert get_in(['purchase'], transaction) == {'items': ['Apple', 'Orange'], 'costs': [0.50, 1.25]}

	def test_update_in(self) -> None:
		D, kw = (self.D, self.kw)
		assert update_in(D({'a': 0}), ['a'], inc, **kw) == D({'a': 1})
		assert update_in(D({'a': 0, 'b': 1}), ['b'], str, **kw) == D({'a': 0, 'b': '1'})
		assert update_in(D({'t': 1, 'v': D({'a': 0})}), ['v', 'a'], inc, **kw) == D({'t': 1, 'v': D({'a': 1})})
		assert update_in(D({}), ['z'], str, None, **kw) == D({'z': 'None'})
		assert update_in(D({}), ['z'], inc, 0, **kw) == D({'z': 1})
		assert update_in(D({}), ['z'], lambda x: x + 'ar', default='b', **kw) == D({'z': 'bar'})
		assert update_in(D({}), [0, 1], inc, default=-1, **kw) == D({0: D({1: 0})})
		assert update_in(D({}), [0, 1], str, default=100, **kw) == D({0: D({1: '100'})})
		assert update_in(D({'foo': 'bar', 1: 50}), ['d', 1, 0], str, 20, **kw) == D({'foo': 'bar', 1: 50, 'd': D({1: D({0: '20'})})})
		# Original mapping is not mutated
		d = D({'x': 1})
		oldd = d
		d2 = update_in(d, ['x'], inc, **kw)
		assert d is oldd
		assert d2 is not oldd
		# Sibling keys at outer level are preserved
		assert update_in(D({'a': 1, 'b': 2}), ['a'], inc, **kw) == D({'a': 2, 'b': 2})
		# Existing nested mapping is not mutated
		inner = D({'b': 1})
		d3 = D({'a': inner})
		d4 = update_in(d3, ['a', 'b'], inc, **kw)
		assert inner['b'] == 1
		assert d4['a']['b'] == 2

	@pytest.mark.parametrize(
		('dictsDefinition', 'expectedMappingDefinition'),
		(
			pytest.param(
				[{2: 'north', 5: 'south'}, {3: 'east', 7: 'west'}],
				{2: 'north', 3: 'east', 5: 'south', 7: 'west'},
				id='non-overlapping keys from both dicts are preserved',
			),
			pytest.param(
				[{2: 'north', 5: 'south'}, {5: 'southeast'}],
				{2: 'north', 5: 'southeast'},
				id='last dict value takes precedence for repeated key',
			),
			pytest.param(
				[{}, {3: 'east'}],
				{3: 'east'},
				id='leading empty dict does not alter result',
			),
			pytest.param(
				[{2: 'north'}, {}],
				{2: 'north'},
				id='trailing empty dict does not alter result',
			),
		),
	)
	def test_mergeProducesExpectedResult(
		self,
		dictsDefinition: list[dict[Any, Any]],
		expectedMappingDefinition: dict[Any, Any],
	) -> None:
		mappingFactory: Callable[[dict[Any, Any]], Any] = self.D
		factoryKeywordArguments: dict[str, Any] = self.kw
		sourceMappings: list[Any] = [mappingFactory(d) for d in dictsDefinition]
		expectedMapping: Any = mappingFactory(expectedMappingDefinition)

		resultMapping: Any = merge(*sourceMappings, **factoryKeywordArguments)

		assert resultMapping == expectedMapping, (
			f'merge returned {resultMapping}, expected {expectedMapping} for inputs {dictsDefinition}.'
		)
		assert resultMapping is not sourceMappings[0], (
			f'merge returned the first source mapping instead of a new mapping for {dictsDefinition}.'
		)
		for sourceMapping, originalDefinition in zip(sourceMappings, dictsDefinition, strict=True):
			assert sourceMapping == mappingFactory(originalDefinition), (
				f'merge mutated source mapping {sourceMapping}, expected {originalDefinition}.'
			)

	@pytest.mark.parametrize(
		('aggregationFunc', 'dictsDefinition', 'expectedMappingDefinition'),
		(
			pytest.param(
				max,
				[{2: 3, 5: 7}, {2: 11, 5: 2}],
				{2: 11, 5: 7},
				id='max selects largest value per key',
			),
			pytest.param(
				min,
				[{2: 3, 5: 7}, {2: 11, 5: 2}],
				{2: 3, 5: 2},
				id='min selects smallest value per key',
			),
			pytest.param(
				sum,
				[{2: 3, 5: 7}, {2: 11}, {5: 13}],
				{2: 14, 5: 20},
				id='sum adds values across three dicts including keys present in only a subset',
			),
		),
	)
	def test_mergeWithAggregatesValues(
		self,
		aggregationFunc: Callable[[list[Any]], Any],
		dictsDefinition: list[dict[Any, Any]],
		expectedMappingDefinition: dict[Any, Any],
	) -> None:
		mappingFactory: Callable[[dict[Any, Any]], Any] = self.D
		factoryKeywordArguments: dict[str, Any] = self.kw
		sourceMappings: list[Any] = [mappingFactory(d) for d in dictsDefinition]
		expectedMapping: Any = mappingFactory(expectedMappingDefinition)

		resultMapping: Any = merge_with(aggregationFunc, *sourceMappings, **factoryKeywordArguments)

		assert resultMapping == expectedMapping, (
			f'merge_with returned {resultMapping}, expected {expectedMapping} for inputs {dictsDefinition}.'
		)
		for sourceMapping, originalDefinition in zip(sourceMappings, dictsDefinition, strict=True):
			assert sourceMapping == mappingFactory(originalDefinition), (
				f'merge_with mutated source mapping {sourceMapping}, expected {originalDefinition}.'
			)

	@pytest.mark.parametrize(
		('sourceMappingDefinition', 'keysPath', 'valueToAssoc', 'expectedMappingDefinition'),
		(
			pytest.param(
				{'alpha': 13, 'beta': 21},
				['alpha'],
				34,
				{'alpha': 34, 'beta': 21},
				id='replaces value at single-key path while preserving siblings',
			),
			pytest.param(
				{'alpha': 21, 'gamma': 55},
				['beta'],
				34,
				{'alpha': 21, 'beta': 34, 'gamma': 55},
				id='inserts new key without altering existing keys',
			),
			pytest.param(
				{},
				['alpha'],
				144,
				{'alpha': 144},
				id='creates key in empty mapping',
			),
		),
	)
	def test_assocInUpdatesAtPath(
		self,
		sourceMappingDefinition: dict[str, Any],
		keysPath: list[str],
		valueToAssoc: Any,
		expectedMappingDefinition: dict[str, Any],
	) -> None:
		mappingFactory: Callable[[dict[Any, Any]], Any] = self.D
		factoryKeywordArguments: dict[str, Any] = self.kw
		sourceMapping: Any = mappingFactory(sourceMappingDefinition)
		expectedMapping: Any = mappingFactory(expectedMappingDefinition)

		resultMapping: Any = assoc_in(sourceMapping, keysPath, valueToAssoc, **factoryKeywordArguments)

		assert resultMapping == expectedMapping, (
			f'assoc_in returned {resultMapping}, expected {expectedMapping} '
			f'for keys={keysPath!r}, value={valueToAssoc!r}, source={sourceMappingDefinition}.'
		)
		assert sourceMapping == mappingFactory(sourceMappingDefinition), (
			f'assoc_in mutated source mapping for keys={keysPath!r}, value={valueToAssoc!r}.'
		)
		assert resultMapping is not sourceMapping, (
			f'assoc_in returned the source mapping instead of a new mapping for keys={keysPath!r}.'
		)

	@pytest.mark.parametrize(
		('sourceMappingDefinition', 'keysPath', 'funcToApply', 'defaultValue', 'expectedMappingDefinition'),
		(
			pytest.param(
				{'alpha': 13, 'beta': 21},
				['alpha'],
				inc,
				None,
				{'alpha': 14, 'beta': 21},
				id='applies func to existing value and preserves siblings',
			),
			pytest.param(
				{},
				['alpha'],
				inc,
				21,
				{'alpha': 22},
				id='applies func to default when key is absent',
			),
			pytest.param(
				{'alpha': 13, 'beta': 21},
				['beta'],
				str,
				None,
				{'alpha': 13, 'beta': '21'},
				id='applies str to value at named key',
			),
			pytest.param(
				{'alpha': 34},
				['gamma'],
				inc,
				55,
				{'alpha': 34, 'gamma': 56},
				id='applies func to default for absent key while preserving existing keys',
			),
		),
	)
	def test_updateInAppliesFuncAtPath(
		self,
		sourceMappingDefinition: dict[str, Any],
		keysPath: list[str],
		funcToApply: Callable[[Any], Any],
		defaultValue: Any,
		expectedMappingDefinition: dict[str, Any],
	) -> None:
		mappingFactory: Callable[[dict[Any, Any]], Any] = self.D
		factoryKeywordArguments: dict[str, Any] = self.kw
		sourceMapping: Any = mappingFactory(sourceMappingDefinition)
		expectedMapping: Any = mappingFactory(expectedMappingDefinition)

		resultMapping: Any = update_in(sourceMapping, keysPath, funcToApply, defaultValue, **factoryKeywordArguments) # pyright: ignore[reportUnknownVariableType]

		assert resultMapping == expectedMapping, (
			f'update_in returned {resultMapping}, expected {expectedMapping} '
			f'for keys={keysPath!r}, func={funcToApply.__name__!r}, default={defaultValue!r}, '  # ty:ignore[unresolved-attribute]
			f'source={sourceMappingDefinition}.'
		)
		assert sourceMapping == mappingFactory(sourceMappingDefinition), (
			f'update_in mutated source mapping for keys={keysPath!r}, func={funcToApply.__name__!r}.'  # ty:ignore[unresolved-attribute]
		)
		assert resultMapping is not sourceMapping, (
			f'update_in returned the source mapping instead of a new mapping for keys={keysPath!r}.'
		)

	def test_factory(self) -> None:
		D, _kw = (self.D, self.kw)
		assert merge(defaultdict(int, D({1: 2})), D({2: 3})) == {1: 2, 2: 3}
		assert merge(defaultdict(int, D({1: 2})), D({2: 3}), factory=lambda: defaultdict(int)) == defaultdict(int, D({1: 2, 2: 3}))
		assert not merge(defaultdict(int, D({1: 2})), D({2: 3}), factory=lambda: defaultdict(int)) == {1: 2, 2: 3}
		assert raises(TypeError, lambda: merge(D({1: 2}), D({2: 3}), factoryy=dict))  # pyright: ignore[reportCallIssue, reportUnknownLambdaType]  # ty:ignore[no-matching-overload]

class defaultdict[KeyType, ValueType](_defaultdict[KeyType, ValueType]):

	__hash__ = None  # pyright: ignore[reportAssignmentType]

	def __eq__(self, other: object) -> bool:
		return super().__eq__(other) and isinstance(other, _defaultdict) and (self.default_factory == other.default_factory)

class TestDefaultDict(TestDict):
	"""Test defaultdict as input and factory

	Class attributes:
		D: callable that inputs a dict and creates or returns a MutableMapping
		kw: kwargs dict to specify "factory" keyword (if applicable)
	"""

	@staticmethod
	def D(dict_: dict[Any, Any]) -> defaultdict[Any, Any]:
		return defaultdict(int, dict_)
	kw: ClassVar[dict[str, Callable[[], defaultdict[Any, Any]]]] = {'factory': makeDefaultDictFactory}

class CustomMapping:
	"""Define methods of the MutableMapping protocol required by dicttoolz"""

	__hash__ = None  # pyright: ignore[reportAssignmentType]

	def __init__(self, *args: Any, **kwargs: Any) -> None:
		self._d = dict(*args, **kwargs)

	def __getitem__(self, key: Any) -> Any:
		return self._d[key]

	def __setitem__(self, key: Any, val: Any) -> None:
		self._d[key] = val

	def __delitem__(self, key: Any) -> None:
		del self._d[key]

	def __iter__(self) -> Iterator[Any]:
		return iter(self._d)

	def __len__(self) -> int:
		return len(self._d)

	def __contains__(self, key: object) -> bool:
		return key in self._d

	def __eq__(self, other: object) -> bool:
		return isinstance(other, CustomMapping) and self._d == other._d

	def __ne__(self, other: object) -> bool:
		return not isinstance(other, CustomMapping) or self._d != other._d

	def keys(self) -> KeysView[Any]:
		return self._d.keys()

	def values(self) -> ValuesView[Any]:
		return self._d.values()

	def items(self) -> ItemsView[Any, Any]:
		return self._d.items()

	def update(self, *args: Any, **kwargs: Any) -> None:
		self._d.update(*args, **kwargs)

	def get(self, key, *args):
		return self._d.get(key, *args)

class TestCustomMapping(TestDict):
	"""Test CustomMapping as input and factory

	Class attributes:
		D: callable that inputs a dict and creates or returns a MutableMapping
		kw: kwargs dict to specify "factory" keyword (if applicable)
	"""
	D: ClassVar[Callable[..., Any]] = CustomMapping
	kw: ClassVar[dict[str, Any]] = {'factory': makeCustomMappingFactory}

@pytest.mark.parametrize(
	('keys', 'coll', 'expectedValue'),
	(
		pytest.param(
			['alpha'],
			{'alpha': 13, 'beta': 21},
			13,
			id='retrieves value at single-key path',
		),
		pytest.param(
			['alpha', 'beta'],
			{'alpha': {'beta': 34}, 'gamma': 55},
			34,
			id='retrieves value at two-key nested path',
		),
		pytest.param(
			['alpha', 'beta', 'gamma'],
			{'alpha': {'beta': {'gamma': 89}, 'delta': 144}, 'epsilon': 233},
			89,
			id='retrieves value at three-key nested path',
		),
		pytest.param(
			[2],
			[3, 5, 7, 11, 13],
			7,
			id='retrieves value at integer index in list',
		),
	),
)
def test_get_in_retrieves_value_at_path(
	keys: list[Any],
	coll: Any,
	expectedValue: Any,
) -> None:
	result: Any = get_in(keys, coll)
	assert result == expectedValue, (
		f'get_in returned {result!r}, expected {expectedValue!r} for keys={keys!r}.'
	)

@pytest.mark.parametrize(
	('keys', 'coll', 'defaultValue', 'expectedValue'),
	(
		pytest.param(
			['alpha'],
			{},
			13,
			13,
			id='returns default for missing single key',
		),
		pytest.param(
			['alpha', 'beta'],
			{'alpha': {}},
			21,
			21,
			id='returns default when intermediate mapping lacks final key',
		),
		pytest.param(
			['alpha', 'beta'],
			{},
			34,
			34,
			id='returns default when first key absent in two-key path',
		),
		pytest.param(
			['alpha'],
			{'beta': 55},
			None,
			None,
			id='returns None when key absent and no default provided',
		),
	),
)
def test_get_in_returns_default_for_missing_path(
	keys: list[Any],
	coll: dict[str, Any],
	defaultValue: Any,
	expectedValue: Any,
) -> None:
	result: Any = get_in(keys, coll, defaultValue)
	assert result == expectedValue, (
		f'get_in returned {result!r}, expected {expectedValue!r} '
		f'for keys={keys!r} with default={defaultValue!r}.'
	)

@pytest.mark.parametrize(
	('sourceMappingDefinition', 'keysPath', 'valueToAssoc', 'expectedMappingDefinition'),
	(
		pytest.param(
			{},
			['alpha', 'beta'],
			21,
			{'alpha': {'beta': 21}},
			id='creates two-level nested path in empty mapping',
		),
		pytest.param(
			{},
			['alpha', 'beta', 'gamma'],
			34,
			{'alpha': {'beta': {'gamma': 34}}},
			id='creates three-level nested path in empty mapping',
		),
		pytest.param(
			{'alpha': {'beta': 55, 'gamma': 89}},
			['alpha', 'beta'],
			144,
			{'alpha': {'beta': 144, 'gamma': 89}},
			id='replaces value at two-key path while preserving nested siblings',
		),
	),
)
def test_assoc_in_nested_paths(
	sourceMappingDefinition: dict[str, Any],
	keysPath: list[str],
	valueToAssoc: Any,
	expectedMappingDefinition: dict[str, Any],
) -> None:
	result: Any = assoc_in(sourceMappingDefinition, keysPath, valueToAssoc)
	assert result == expectedMappingDefinition, (
		f'assoc_in returned {result}, expected {expectedMappingDefinition} '
		f'for keys={keysPath!r}, value={valueToAssoc!r}, source={sourceMappingDefinition}.'
	)
	assert result is not sourceMappingDefinition, (
		f'assoc_in returned the source mapping instead of a new mapping for keys={keysPath!r}.'
	)

@pytest.mark.parametrize(
	('sourceMappingDefinition', 'keysPath', 'funcToApply', 'defaultValue', 'expectedMappingDefinition'),
	(
		pytest.param(
			{'alpha': {'beta': 34}},
			['alpha', 'beta'],
			inc,
			None,
			{'alpha': {'beta': 35}},
			id='applies func at two-level nested path to existing value',
		),
		pytest.param(
			{},
			['alpha', 'beta'],
			inc,
			21,
			{'alpha': {'beta': 22}},
			id='creates two-level nested path and applies func to default',
		),
		pytest.param(
			{'alpha': {'beta': 55, 'gamma': 89}},
			['alpha', 'beta'],
			str,
			None,
			{'alpha': {'beta': '55', 'gamma': 89}},
			id='applies func at two-level path while preserving nested siblings',
		),
	),
)
def test_update_in_nested_paths(
	sourceMappingDefinition: dict[str, Any],
	keysPath: list[str],
	funcToApply: Callable[[Any], Any],
	defaultValue: Any,
	expectedMappingDefinition: dict[str, Any],
) -> None:
	result: Any = update_in(sourceMappingDefinition, keysPath, funcToApply, defaultValue)
	assert result == expectedMappingDefinition, (
		f'update_in returned {result}, expected {expectedMappingDefinition} '
		f'for keys={keysPath!r}, func={funcToApply.__name__!r}, default={defaultValue!r}, '  # ty:ignore[unresolved-attribute]
		f'source={sourceMappingDefinition}.'
	)
	assert result is not sourceMappingDefinition, (
		f'update_in returned the source dict instead of a new mapping for keys={keysPath!r}.'
	)

def test_get_in_no_default() -> None:
	with pytest.raises(KeyError):
		get_in(['y'], {}, no_default=True)
	with pytest.raises(IndexError):
		get_in([0], [], no_default=True)  # pyright: ignore[reportArgumentType]  # ty:ignore[no-matching-overload]
	with pytest.raises(TypeError):
		get_in(['x', 0], {'x': 42}, no_default=True)  # ty:ignore[no-matching-overload]

def test_get_in_empty_keys() -> None:
	coll: dict[str, int] = {'a': 1}
	assert get_in([], coll) is coll

def test_environ() -> None:
	assert keymap(identity, os.environ) == os.environ
	assert valmap(identity, os.environ) == os.environ
	assert itemmap(identity, os.environ) == os.environ

def test_merge_with_non_dict_mappings() -> None:

	class Foo(Mapping[int, int]):

		def __init__(self, d: dict[int, int]) -> None:
			self.d: dict[int, int] = d

		def __iter__(self) -> Iterator[int]:
			return iter(self.d)

		def __getitem__(self, key: int) -> int:
			return self.d[key]

		def __len__(self) -> int:
			return len(self.d)

	d = Foo({1: 1})
	assert merge(d) is d or merge(d) == {1: 1}
	assert merge_with(sum, d) == {1: 1}



"""Tests for data structure utilities.

(AI generated docstring)

This module validates the behavior of data structure manipulation functions,
string conversion utilities, and Run-Length Encoding (RLE) tools.

"""
# pyright: standard
from collections.abc import Callable, Iterable, Iterator
from decimal import Decimal
from fractions import Fraction
from hunterMakesPy import autoDecodingRLE, stringItUp, updateExtendPolishDictionaryLists
from hunterMakesPy.tests.conftest import standardizedEqualTo
from numpy.typing import NDArray
from typing import Any
import datetime
import numpy
import pytest
import sys

class CustomIterable:
	"""A simple custom iterable for testing purposes.

	(AI generated docstring)

	Attributes
	----------
	items : Iterable[Any]
		The items to iterate over.

	"""
	def __init__(self, items: Iterable[Any]) -> None: self.items: Iterable[Any] = items
	def __iter__(self) -> Iterator[Any]: return iter(self.items)

@pytest.mark.parametrize("description,value_scrapPile,expected", [
	# Basic types and structures
	("Empty input", [], []),
	("Prime numbers", [11, 13, 17], ['11', '13', '17']),
	("Cardinal directions", ["NE", "SW", "SE"], ["NE", "SW", "SE"]),
	("Country codes", ["FR", "JP", "BR"], ["FR", "JP", "BR"]),
	("Boolean values", [True, False], ['True', 'False']),
	("None value", [None], ['None']),
	# Numbers and numeric types
	("Fibonacci floats", [2.584, -4.236, 6.854], ['2.584', '-4.236', '6.854']),
	("Complex with primes", [complex(11,0), complex(13,0)], ['(11+0j)', '(13+0j)']),
	("Decimal and Fraction", [Decimal('3.141'), Fraction(89, 55)], ['3.141', '89/55']),
	("NumPy primes", numpy.array([11, 13, 17]), ['11', '13', '17']),
	# Temporal types with meaningful dates
	("Historical date", [datetime.date(1789, 7, 14)], ['1789-07-14']),  # Bastille Day
	("Time zones", [datetime.time(23, 11, 37)], ['23:11:37']),  # Non-standard time
	("Moon landing", [datetime.datetime(1969, 7, 20, 20, 17, 40)], ['1969-07-20 20:17:40']),
	# Binary data - accepting either representation
	("Prime bytes", [b'\x0B', b'\x0D', b'\x11'], [repr(b'\x0b'), repr(b'\x0d'), repr(b'\x11')]),  # Let Python choose representation
	("Custom bytearray", [bytearray(b"DEADBEEF")], ["bytearray(b'DEADBEEF')"]),
	("Memory view decoded", memoryview(b"DEADBEEF"), ["DEADBEEF"]),
	# Nested structures with unique values
	("Nested dictionary", {'phi': 1.618, 'euler': 2.718}, ['phi', '1.618', 'euler', '2.718']),
	("Mixed nesting", [{'NE': 37}, {'SW': 41}], ['NE', '37', 'SW', '41']),
	("Tuples and lists", [(13, 17), [19, 23]], ['13', '17', '19', '23']),
	("Sets and frozensets", [{37, 41}, frozenset([43, 47])], ['41', '37', '43', '47']),
	# Special cases and error handling
	("NaN and Infinities", [float('nan'), float('inf'), -float('inf')], ['nan', 'inf', '-inf']),
	("Large prime", [10**19 + 33], ['10000000000000000033']),
	("Simple recursive", [[[...]]], ['Ellipsis']),  # Recursive list
	("Complex recursive", {'self': {'self': None}}, ['self', 'self', 'None']),
	# Generators and custom iterables
	("Generator from primes", (x for x in [11, 13, 17]), ['11', '13', '17']),
	("Iterator from Fibonacci", iter([3, 5, 8, 13]), ['3', '5', '8', '13']),
	("Custom iterable cardinal", CustomIterable(["NW", "SE", "NE"]), ["NW", "SE", "NE"]),
	("Custom iterable empty", CustomIterable([]), []),
	# Weird stuff
	("Bad __str__", type('BadStr', (), {'__str__': lambda x: None})(), [None]),
	# Error cases
	("Raising __str__", type('RaisingStr', (), {'__str__': lambda x: 1/0})(), ZeroDivisionError),
], ids=lambda x: x if isinstance(x, str) else "")
def testStringItUp(description: str, value_scrapPile: list[Any], expected: list[str] | type[Exception]) -> None:
	"""Test stringItUp with various inputs.

	Parameters
	----------
	description : str
		Description of the test case.
	value_scrapPile : list[Any]
		List of values to convert to strings.
	expected : list[str] | type[Exception]
		Expected list of string representations or an Exception type.

	"""
	standardizedEqualTo(expected, stringItUp, value_scrapPile)

@pytest.mark.parametrize("description,value_dictionaryLists,keywordArguments,expected", [
	("Mixed value types", ({'ne': [11, 'prime'], 'sw': [True, None]}, {'ne': [3.141, 'golden'], 'sw': [False, 'void']}), {'destroyDuplicates': False, 'reorderLists': False}, {'ne': [11, 'prime', 3.141, 'golden'], 'sw': [True, None, False, 'void']}),
	("Empty dictionaries", (dict[str, list[Any]](), dict[str, list[Any]]()), dict[str, Any](), dict[str, list[Any]]()),
	("Tuple values", ({'ne': (11, 13), 'sw': (17,)}, {'ne': (19, 23, 13, 29, 11), 'sw': (31, 17, 37)}), {'destroyDuplicates': False, 'reorderLists': False}, {'ne': [11, 13, 19, 23, 13, 29, 11], 'sw': [17, 31, 17, 37]}),
	("Set values", ({'ne': {11, 13}, 'sw': {17}}, {'ne': {19, 23, 13, 29, 11}, 'sw': {31, 17, 37}}), {'destroyDuplicates': True, 'reorderLists': True}, {'ne': [11, 13, 19, 23, 29], 'sw': [17, 31, 37]}),
	("NumPy arrays", ({'ne': numpy.array([11, 13]), 'sw': numpy.array([17])}, {'ne': numpy.array([19, 23, 13, 29, 11]), 'sw': numpy.array([31, 17, 37])}), {'destroyDuplicates': False, 'reorderLists': False}, {'ne': [11, 13, 19, 23, 13, 29, 11], 'sw': [17, 31, 17, 37]}),
	("Destroy duplicates", ({'fr': [11, 13], 'jp': [17]}, {'fr': [19, 23, 13, 29, 11], 'jp': [31, 17, 37]}), {'destroyDuplicates': True, 'reorderLists': False}, {'fr': [11, 13, 19, 23, 29], 'jp': [17, 31, 37]}),
	("Non-string keys", ({None: [13], True: [17]}, {19: [23], (29, 31): [37]}), {'destroyDuplicates': False, 'reorderLists': False}, {'None': [13], 'True': [17], '19': [23], '(29, 31)': [37]}),
	("Reorder lists", ({'fr': [11, 13], 'jp': [17]}, {'fr': [19, 23, 13, 29, 11], 'jp': [31, 17, 37]}), {'destroyDuplicates': False, 'reorderLists': True}, {'fr': [11, 11, 13, 13, 19, 23, 29], 'jp': [17, 17, 31, 37]}),
	("Non-iterable values", ({'ne': 13, 'sw': 17}, {'ne': 19, 'nw': 23}), {'destroyDuplicates': False, 'reorderLists': False}, TypeError),
	("Skip erroneous types", ({'ne': [11, 13], 'sw': [17, 19]}, {'ne': 23, 'nw': 29}), {'killErroneousDataTypes': True}, {'ne': [11, 13], 'sw': [17, 19]}),
], ids=lambda x: x if isinstance(x, str) else "")
def testUpdateExtendPolishDictionaryLists(description: str, value_dictionaryLists: tuple[dict[str, Any], ...], keywordArguments: dict[str, Any], expected: dict[str, Any] | type[TypeError]) -> None:
	"""Test dictionary list updating and extension logic.

	(AI generated docstring)

	Parameters
	----------
	description : str
		Description of the test case.
	value_dictionaryLists : tuple[dict[str, Any], ...]
		Tuple of dictionaries to merge/update.
	keywordArguments : dict[str, Any]
		Keyword arguments controlling the behavior (e.g., destroyDuplicates).
	expected : dict[str, Any] | type[TypeError]
		The expected result dictionary or an exception type.

	"""
	standardizedEqualTo(expected, updateExtendPolishDictionaryLists, *value_dictionaryLists, **keywordArguments)

# ruff: noqa: RUF005

def generateCartesianMapping(dimensions: tuple[int, int], formula: Callable[[int, int], int]) -> NDArray[numpy.int32]:
	"""Generate a 2D cartesian mapping based on a formula.

	Parameters
	----------
	dimensions : tuple[int, int]
		The (height, width) of the array to generate.
	formula : Callable[[int, int], int]
		A function taking (x, y) coordinates and returning an integer value.

	Returns
	-------
	arrayMapping : NDArray[numpy.int32]
		The generated 2D numpy array.

	"""
	height, width = dimensions
	arrayMapping = numpy.zeros((height, width), dtype=numpy.int32)

	for y in range(height):
		for x in range(width):
			arrayMapping[y, x] = formula(x, y)

	return arrayMapping

def generateWavePattern(dimensions: tuple[int, int]) -> NDArray[numpy.int32]:
	"""Generate a sine wave pattern that produces many RLE-friendly sequences.

	Parameters
	----------
	dimensions : tuple[int, int]
		The (height, width) of the array.

	Returns
	-------
	arrayMapping : NDArray[numpy.int32]
		The generated wave pattern array.

	"""

	def waveFormula(x: int, y: int) -> int:
		return int(10 * numpy.sin(x / 5) + 10 * numpy.sin(y / 5))

	return generateCartesianMapping(dimensions, waveFormula)

def generateChessboard(dimensions: tuple[int, int], squareSize: int = 4) -> NDArray[numpy.int32]:
	"""Generate a chessboard pattern with alternating values.

	Parameters
	----------
	dimensions : tuple[int, int]
		The (height, width) of the array.
	squareSize : int = 4
		The size of each chessboard square.

	Returns
	-------
	arrayMapping : NDArray[numpy.int32]
		The generated chessboard pattern array.

	"""
	def chessboardFormula(x: int, y: int) -> int:
		return 1 if ((x // squareSize) + (y // squareSize)) % 2 == 0 else 0

	return generateCartesianMapping(dimensions, chessboardFormula)

def generatePrimeModuloMatrix(dimensions: tuple[int, int], modulus: int = 6) -> NDArray[numpy.int32]:
	"""Generate a matrix where each cell is (x*y) % modulus.

	Parameters
	----------
	dimensions : tuple[int, int]
		The (height, width) of the array.
	modulus : int = 6
		The modulus to apply to the product of coordinates.

	Returns
	-------
	arrayMapping : NDArray[numpy.int32]
		The generated prime modulo matrix.

	"""
	def primeModuloFormula(x: int, y: int) -> int:
		return ((x + 1) * (y + 1)) % modulus

	return generateCartesianMapping(dimensions, primeModuloFormula)

def generateSpiralPattern(dimensions: tuple[int, int], scale: int = 1) -> NDArray[numpy.int32]:
	"""Generate a spiral pattern that creates interesting RLE sequences.

	Parameters
	----------
	dimensions : tuple[int, int]
		The (height, width) of the array.
	scale : int = 1
		Scaling factor for the spiral values.

	Returns
	-------
	arrayMapping : NDArray[numpy.int32]
		The generated spiral pattern array.

	"""
	height, width = dimensions

	def spiralFormula(x: int, y: int) -> int:
		xCenter: int = width // 2
		yCenter: int = height // 2
		distanceX: int = x - xCenter
		distanceY: int = y - yCenter
		distance: float = numpy.sqrt(distanceX**2 + distanceY**2)
		angle: float = numpy.arctan2(distanceY, distanceX)
		return int((distance + 5 * angle) * scale) % 10

	return generateCartesianMapping(dimensions, spiralFormula)

def generateSignedQuadraticFunction(dimensions: tuple[int, int]) -> NDArray[numpy.int32]:
	"""Generate a matrix with a quadratic function that includes negative values.

	Parameters
	----------
	dimensions : tuple[int, int]
		The (height, width) of the array.

	Returns
	-------
	arrayMapping : NDArray[numpy.int32]
		The generated quadratic function array.

	"""
	height, width = dimensions

	def quadraticFormula(x: int, y: int) -> int:
		xCenter: int = width // 2
		yCenter: int = height // 2
		return (x - xCenter)**2 - (y - yCenter)**2

	return generateCartesianMapping(dimensions, quadraticFormula)

def generateTilePattern(dimensions: tuple[int, int], tileSize: int = 10) -> NDArray[numpy.int32]:
	"""Generate a repeating tile pattern.

	Parameters
	----------
	dimensions : tuple[int, int]
		The (height, width) of the array.
	tileSize : int = 10
		The size of the repeating tile.

	Returns
	-------
	arrayMapping : NDArray[numpy.int32]
		The generated tile pattern array.

	"""
	def tileFormula(x: int, y: int) -> int:
		patternX: int = x % tileSize
		patternY: int = y % tileSize
		if patternX < patternY:
			return patternX
		else:
			return patternY

	return generateCartesianMapping(dimensions, tileFormula)

def generateRepeatingZones(dimensions: tuple[int, int], zoneWidth: int = 15) -> NDArray[numpy.int32]:
	"""Generate horizontal zones with repeating values.

	Parameters
	----------
	dimensions : tuple[int, int]
		The (height, width) of the array.
	zoneWidth : int = 15
		The width of each zone.

	Returns
	-------
	arrayMapping : NDArray[numpy.int32]
		The generated zonal pattern array.

	"""
	def zoneFormula(x: int, y: int) -> int:
		zone: int = y // zoneWidth
		return zone % 5  # 5 different zones

	return generateCartesianMapping(dimensions, zoneFormula)

def generateStepPattern(dimensions: tuple[int, int], step: int = 5) -> NDArray[numpy.int32]:
	"""Generate a stepping pattern that increases along the x-axis.

	Parameters
	----------
	dimensions : tuple[int, int]
		The (height, width) of the array.
	step : int = 5
		The step size for value increments.

	Returns
	-------
	arrayMapping : NDArray[numpy.int32]
		The generated step pattern array.

	"""
	def stepFormula(x: int, y: int) -> int:
		return x // step

	return generateCartesianMapping(dimensions, stepFormula)

def generateAlternatingColumns(dimensions: tuple[int, int], blockSize: int = 1) -> NDArray[numpy.int32]:
	"""Generate alternating columns with different values.

	Parameters
	----------
	dimensions : tuple[int, int]
		The (height, width) of the array.
	blockSize : int = 1
		The width of each column block.

	Returns
	-------
	arrayMapping : NDArray[numpy.int32]
		The generated alternating columns array.

	"""
	def columnFormula(x: int, y: int) -> int:
		return (x // blockSize) % 2

	return generateCartesianMapping(dimensions, columnFormula)

AUTO_DECODING_RLE_EXACT_CASES: list[tuple[str, NDArray[numpy.integer[Any]], str]] = [
	("One range", numpy.array(list(range(50,60))), "[*range(50,60)]"),
	("Value, range", numpy.array([123]+list(range(71,81))), "[123,*range(71,81)]"),
	("range, value", numpy.array(list(range(91,97))+[101]), "[*range(91,97),101]"),
	("Value, range, value", numpy.array([151]+list(range(163,171))+[181]), "[151,*range(163,171),181]"),
	("Repeat values", numpy.array([191, 191, 191]), "[191]*3"),
	("Value with repeat", numpy.array([211, 223, 223, 223]), "[211]+[223]*3"),
	("Range with repeat", numpy.array(list(range(251,257))+[271, 271, 271]), "[*range(251,257)]+[271]*3"),
	("Value, range, repeat", numpy.array([281]+list(range(291,297))+[307, 307]), "[281,*range(291,297)]+[307]*2"),
	("repeat, value", numpy.array([313, 313, 313, 331, 331, 349]), "[313]*3+[331]*2+[349]"),
	("repeat, range", numpy.array([373, 373, 373]+list(range(383,389))), "[373]*3+[*range(383,389)]"),
	("repeat, range, value", numpy.array(7*[401]+list(range(409,415))+[421]), "[401]*7+[*range(409,415),421]"),
	("Repeated primes", numpy.array([431, 431, 431, 443, 443, 457]), "[431]*3+[443]*2+[457]"),
	("Two Ranges", numpy.array(list(range(461,471))+list(range(479,487))), "[*range(461,471),*range(479,487)]"),
	("2D array primes", numpy.array([[491, 499, 503], [509, 521, 523]]), "[[491,499,503],[509,521,523]]"),
	("3D array primes", numpy.array([[[541, 547], [557, 563]], [[569, 571], [577, 587]]]), "[[[541,547],[557,563]],[[569,571],[577,587]]]"),
]

AUTO_DECODING_RLE_REALISTIC_CASES: list[tuple[str, NDArray[numpy.integer[Any]]]] = [
	("Simple range", numpy.array(list(range(50,60)))),
	("Small chessboard", generateChessboard((8, 8))),
	("Alternating columns", generateAlternatingColumns((5, 20), 2)),
	("Step pattern", generateStepPattern((6, 30), 3)),
	("Repeating zones", generateRepeatingZones((40, 40), 8)),
	("Tile pattern", generateTilePattern((15, 15), 5)),
	("Signed quadratic", generateSignedQuadraticFunction((10, 10))),
	("Prime modulo", generatePrimeModuloMatrix((12, 12), 7)),
	("Wave pattern", generateWavePattern((20, 20))),
	("Spiral pattern", generateSpiralPattern((15, 15), 2)),
]

AUTO_DECODING_RLE_ROUNDTRIP_CASES: list[tuple[str, NDArray[numpy.integer[Any]]]] = [
	("1D range", numpy.array(list(range(50,60)))),
	("1D repeats", numpy.array([5, 5, 5, 10, 10, 10, 10, 20])),
	("1D mixed", numpy.array([1, 3, 5, 6, 7, 8, 9, 2, 2, 2])),
	("2D small", numpy.array([[1, 2, 3], [4, 5, 6]])),
	("2D chessboard", generateChessboard((8, 8))),
	("2D zones", generateRepeatingZones((10, 10), 3)),
	("3D tiny", numpy.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])),
	("1D with zero start", numpy.array(list(range(10)))),
	("1D negative values", numpy.array([-5, -4, -3, -2, -1, 0, 1, 2])),
	("1D single element", numpy.array([42])),
	("Large cartesian", generateCartesianMapping((50, 50), lambda x, y: ((x * 17) % 11 + (y * 13) % 7) % 10)),
]

@pytest.mark.parametrize("description,value_arrayTarget,expected", AUTO_DECODING_RLE_EXACT_CASES, ids=lambda x: x if isinstance(x, str) else "")
def testAutoDecodingRLE(description: str, value_arrayTarget: NDArray[numpy.integer[Any]], expected: str) -> None:
	"""Test autoDecodingRLE with various input arrays.

	Parameters
	----------
	description : str
		Description of the test case.
	value_arrayTarget : NDArray[numpy.integer[Any]]
		The input numpy array to encode.
	expected : str
		The expected string representation of the RLE.

	"""
	standardizedEqualTo(expected, autoDecodingRLE, value_arrayTarget)

@pytest.mark.parametrize("description,value_arrayTarget", AUTO_DECODING_RLE_REALISTIC_CASES, ids=lambda x: x if isinstance(x, str) else "")
def testAutoDecodingRLEWithRealisticData(description: str, value_arrayTarget: NDArray[numpy.integer[Any]]) -> None:
	"""Test autoDecodingRLE with more realistic data patterns.

	Parameters
	----------
	description : str
		Description of the test pattern.
	value_arrayTarget : NDArray[numpy.integer[Any]]
		The input array generated by a pattern function.

	"""
	# Here we test the function behavior rather than expected string output
	resultRLE: str = autoDecodingRLE(value_arrayTarget)

	# Test that the result is a valid string
	assert isinstance(resultRLE, str), (
		f"autoDecodingRLE returned {type(resultRLE).__name__}, expected str for {description=}."
	)

	# Test that the result contains the expected syntax elements
	assert "[" in resultRLE, (
		f"autoDecodingRLE returned {resultRLE!r}, expected output containing '[' for {description=}."
	)
	assert "]" in resultRLE, (
		f"autoDecodingRLE returned {resultRLE!r}, expected output containing ']' for {description=}."
	)

	# Check that the result is more compact than the raw string representation
	rawStrLength: int = len(str(value_arrayTarget.tolist()))
	encodedLength: int = len(resultRLE)
	assert encodedLength <= rawStrLength, (
		f"autoDecodingRLE produced length {encodedLength}, expected <= {rawStrLength} for {description=}."
	)

@pytest.mark.parametrize("description,addSpaces", [
	("With spaces", True),
	("Without spaces", False),
], ids=lambda x: x if isinstance(x, str) else "")
def testAutoDecodingRLEWithSpaces(description: str, addSpaces: bool) -> None:
	"""Test that the addSpaces parameter affects the internal comparison logic.

	Note: addSpaces doesn't directly change the output format, it just changes
	the comparison when measuring the length of the string representation.
	The feature exists because `ast` inserts spaces in its string representation.

	Parameters
	----------
	description : str
		Description of the test configuration.
	addSpaces : bool
		Value for the `assumeAddSpaces` parameter.

	"""
	# Create a pattern that has repeated sequences to trigger the RLE logic
	arrayTarget: NDArray[numpy.int32] = generateRepeatingZones((10, 10), 2)

	# Test both configurations
	resultWithSpacesFlag: str = autoDecodingRLE(arrayTarget, assumeAddSpaces=addSpaces)
	resultNoSpacesFlag: str = autoDecodingRLE(arrayTarget, assumeAddSpaces=False)

	assert isinstance(resultWithSpacesFlag, str), (
		f"autoDecodingRLE returned {type(resultWithSpacesFlag).__name__}, expected str for {description=} and {addSpaces=}."
	)
	assert isinstance(resultNoSpacesFlag, str), (
		f"autoDecodingRLE returned {type(resultNoSpacesFlag).__name__}, expected str for baseline comparison with {description=}."
	)

@pytest.mark.parametrize("description,dimensions", [
	("Large Cartesian mapping", (100, 100)),
], ids=lambda x: x if isinstance(x, str) else "")
def testAutoDecodingRLELargeCartesianMapping(description: str, dimensions: tuple[int, int]) -> None:
	"""Test autoDecodingRLE with a large (100x100) cartesian mapping."""

	# Generate a large cartesian mapping with a complex pattern
	def complexFormula(x: int, y: int) -> int:
		return ((x * 17) % 11 + (y * 13) % 7) % 10

	arrayMapping: NDArray[numpy.int32] = generateCartesianMapping(dimensions, complexFormula)

	# Verify the function works with large arrays
	resultRLE: str = autoDecodingRLE(arrayMapping)

	# The result should be a valid string representation
	assert isinstance(resultRLE, str), (
		f"autoDecodingRLE returned {type(resultRLE).__name__}, expected str for {description=}."
	)
	assert "[" in resultRLE, (
		f"autoDecodingRLE returned {resultRLE!r}, expected output containing '[' for {description=}."
	)
	assert "]" in resultRLE, (
		f"autoDecodingRLE returned {resultRLE!r}, expected output containing ']' for {description=}."
	)

	# The RLE encoding should be more compact than the raw representation
	rawStrLength: int = len(str(arrayMapping.tolist()))
	encodedLength: int = len(resultRLE)
	assert encodedLength <= rawStrLength, (
		f"autoDecodingRLE produced length {encodedLength}, expected <= {rawStrLength} for {description=}."
	)

@pytest.mark.parametrize("description,value_arrayTarget", AUTO_DECODING_RLE_ROUNDTRIP_CASES, ids=lambda x: x if isinstance(x, str) else "")
def testAutoDecodingRLERoundtrip(description: str, value_arrayTarget: NDArray[numpy.integer[Any]]) -> None:
	"""Test that eval of autoDecodingRLE output reproduces the original array data.

	Parameters
	----------
	description : str
		Description of the test case.
	value_arrayTarget : NDArray[numpy.integer[Any]]
		The input array to encode and then roundtrip via eval.

	"""
	resultRLE: str = autoDecodingRLE(value_arrayTarget)
	decodedData = eval(resultRLE)  # noqa: S307
	reconstructedArray: NDArray[numpy.integer[Any]] = numpy.array(decodedData)
	numpy.testing.assert_array_equal(reconstructedArray, value_arrayTarget)



"""Internal module for better introspection of builtins.

The main functions are ``is_builtin_valid_args``, ``is_builtin_partial_args``,
and ``has_unknown_args``.  Other functions in this module support these three.

Notably, we create a ``signatures`` registry to enable introspection of
builtin functions in any Python version.  This includes builtins that
have more than one valid signature.  Currently, the registry includes
builtins from ``builtins``, ``functools``, ``itertools``, and ``operator``
modules.  More can be added as requested.  We don't guarantee full coverage.

Everything in this module should be regarded as implementation details.
Users should try to not use this module directly.
"""
from .functoolz import has_keywords, has_varargs, is_arity, is_partial_args, num_required_args
from importlib import import_module
import builtins
import functools
import inspect
import itertools
import operator

module_info = {}
module_info[builtins] = dict(abs=[lambda x: None], all=[lambda iterable: None], anext=[lambda aiterator: None, lambda aiterator, default: None], any=[lambda iterable: None], apply=[lambda object: None, lambda object, args: None, lambda object, args, kwargs: None], ascii=[lambda obj: None], bin=[lambda number: None], bool=[lambda x=False: None], buffer=[lambda object: None, lambda object, offset: None, lambda object, offset, size: None], bytearray=[lambda: None, lambda int: None, lambda string, encoding='utf8', errors='strict': None], callable=[lambda obj: None], chr=[lambda i: None], classmethod=[lambda function: None], cmp=[lambda x, y: None], coerce=[lambda x, y: None], complex=[lambda real=0, imag=0: None], delattr=[lambda obj, name: None], dict=[lambda **kwargs: None, lambda mapping, **kwargs: None], dir=[lambda: None, lambda object: None], divmod=[lambda x, y: None], enumerate=[(0, lambda iterable, start=0: None)], eval=[lambda source: None, lambda source, globals: None, lambda source, globals, locals: None], execfile=[lambda filename: None, lambda filename, globals: None, lambda filename, globals, locals: None], file=[(0, lambda name, mode='r', buffering=-1: None)], filter=[lambda function, iterable: None], float=[lambda x=0.0: None], format=[lambda value: None, lambda value, format_spec: None], frozenset=[lambda: None, lambda iterable: None], getattr=[lambda object, name: None, lambda object, name, default: None], globals=[lambda: None], hasattr=[lambda obj, name: None], hash=[lambda obj: None], hex=[lambda number: None], id=[lambda obj: None], input=[lambda: None, lambda prompt: None], int=[lambda x=0: None, (0, lambda x, base=10: None)], intern=[lambda string: None], isinstance=[lambda obj, class_or_tuple: None], issubclass=[lambda cls, class_or_tuple: None], iter=[lambda iterable: None, lambda callable, sentinel: None], len=[lambda obj: None], list=[lambda: None, lambda iterable: None], locals=[lambda: None], long=[lambda x=0: None, (0, lambda x, base=10: None)], map=[lambda func, sequence, *iterables: None], memoryview=[(0, lambda object: None)], next=[lambda iterator: None, lambda iterator, default: None], object=[lambda: None], oct=[lambda number: None], ord=[lambda c: None], pow=[lambda x, y: None, lambda x, y, z: None], property=[lambda fget=None, fset=None, fdel=None, doc=None: None], range=[lambda stop: None, lambda start, stop: None, lambda start, stop, step: None], raw_input=[lambda: None, lambda prompt: None], reduce=[lambda function, sequence: None, lambda function, sequence, initial: None], reload=[lambda module: None], repr=[lambda obj: None], reversed=[lambda sequence: None], round=[(0, lambda number, ndigits=0: None)], set=[lambda: None, lambda iterable: None], setattr=[lambda obj, name, value: None], slice=[lambda stop: None, lambda start, stop: None, lambda start, stop, step: None], staticmethod=[lambda function: None], sum=[lambda iterable: None, lambda iterable, start: None], super=[lambda type: None, lambda type, obj: None], tuple=[lambda: None, lambda iterable: None], type=[lambda object: None, lambda name, bases, dict: None], unichr=[lambda i: None], unicode=[lambda object: None, lambda string='', encoding='utf8', errors='strict': None], vars=[lambda: None, lambda object: None], xrange=[lambda stop: None, lambda start, stop: None, lambda start, stop, step: None], zip=[lambda *iterables: None], __build_class__=[(2, lambda func, name, *bases, **kwds: None, ('metaclass',))], __import__=[(0, lambda name, globals=None, locals=None, fromlist=None, level=None: None)])
module_info[builtins]['exec'] = [lambda source: None, lambda source, globals: None, lambda source, globals, locals: None]
module_info[builtins].update(breakpoint=[lambda *args, **kws: None], bytes=[lambda: None, lambda int: None, lambda string, encoding='utf8', errors='strict': None], compile=[(0, lambda source, filename, mode, flags=0, dont_inherit=False, optimize=-1: None)], max=[(1, lambda iterable: None, ('default', 'key')), (1, lambda arg1, arg2, *args: None, ('key',))], min=[(1, lambda iterable: None, ('default', 'key')), (1, lambda arg1, arg2, *args: None, ('key',))], open=[(0, lambda file, mode='r', buffering=-1, encoding=None, errors=None, newline=None, closefd=True, opener=None: None)], sorted=[(1, lambda iterable: None, ('key', 'reverse'))], str=[lambda object='', encoding='utf', errors='strict': None])
module_info[builtins]['print'] = [(0, lambda *args: None, ('sep', 'end', 'file', 'flush'))]
module_info[functools] = dict(cmp_to_key=[(0, lambda mycmp: None)], partial=[lambda func, *args, **kwargs: None], partialmethod=[lambda func, *args, **kwargs: None], reduce=[lambda function, sequence: None, lambda function, sequence, initial: None])
module_info[itertools] = dict(accumulate=[(0, lambda iterable, func=None: None)], chain=[lambda *iterables: None], combinations=[(0, lambda iterable, r: None)], combinations_with_replacement=[(0, lambda iterable, r: None)], compress=[(0, lambda data, selectors: None)], count=[lambda start=0, step=1: None], cycle=[lambda iterable: None], dropwhile=[lambda predicate, iterable: None], filterfalse=[lambda function, sequence: None], groupby=[(0, lambda iterable, key=None: None)], ifilter=[lambda function, sequence: None], ifilterfalse=[lambda function, sequence: None], imap=[lambda func, sequence, *iterables: None], islice=[lambda iterable, stop: None, lambda iterable, start, stop: None, lambda iterable, start, stop, step: None], izip=[lambda *iterables: None], izip_longest=[(0, lambda *iterables: None, ('fillvalue',))], pairwise=[lambda iterable: None], permutations=[(0, lambda iterable, r=0: None)], product=[(0, lambda *iterables: None, ('repeat',))], repeat=[(0, lambda object, times=0: None)], starmap=[lambda function, sequence: None], takewhile=[lambda predicate, iterable: None], tee=[lambda iterable: None, lambda iterable, n: None], zip_longest=[(0, lambda *iterables: None, ('fillvalue',))])
module_info[operator] = dict(__abs__=[lambda a: None], __add__=[lambda a, b: None], __and__=[lambda a, b: None], __concat__=[lambda a, b: None], __contains__=[lambda a, b: None], __delitem__=[lambda a, b: None], __delslice__=[lambda a, b, c: None], __div__=[lambda a, b: None], __eq__=[lambda a, b: None], __floordiv__=[lambda a, b: None], __ge__=[lambda a, b: None], __getitem__=[lambda a, b: None], __getslice__=[lambda a, b, c: None], __gt__=[lambda a, b: None], __iadd__=[lambda a, b: None], __iand__=[lambda a, b: None], __iconcat__=[lambda a, b: None], __idiv__=[lambda a, b: None], __ifloordiv__=[lambda a, b: None], __ilshift__=[lambda a, b: None], __imatmul__=[lambda a, b: None], __imod__=[lambda a, b: None], __imul__=[lambda a, b: None], __index__=[lambda a: None], __inv__=[lambda a: None], __invert__=[lambda a: None], __ior__=[lambda a, b: None], __ipow__=[lambda a, b: None], __irepeat__=[lambda a, b: None], __irshift__=[lambda a, b: None], __isub__=[lambda a, b: None], __itruediv__=[lambda a, b: None], __ixor__=[lambda a, b: None], __le__=[lambda a, b: None], __lshift__=[lambda a, b: None], __lt__=[lambda a, b: None], __matmul__=[lambda a, b: None], __mod__=[lambda a, b: None], __mul__=[lambda a, b: None], __ne__=[lambda a, b: None], __neg__=[lambda a: None], __not__=[lambda a: None], __or__=[lambda a, b: None], __pos__=[lambda a: None], __pow__=[lambda a, b: None], __repeat__=[lambda a, b: None], __rshift__=[lambda a, b: None], __setitem__=[lambda a, b, c: None], __setslice__=[lambda a, b, c, d: None], __sub__=[lambda a, b: None], __truediv__=[lambda a, b: None], __xor__=[lambda a, b: None], _abs=[lambda x: None], _compare_digest=[lambda a, b: None], abs=[lambda a: None], add=[lambda a, b: None], and_=[lambda a, b: None], attrgetter=[lambda attr, *args: None], concat=[lambda a, b: None], contains=[lambda a, b: None], countOf=[lambda a, b: None], delitem=[lambda a, b: None], delslice=[lambda a, b, c: None], div=[lambda a, b: None], eq=[lambda a, b: None], floordiv=[lambda a, b: None], ge=[lambda a, b: None], getitem=[lambda a, b: None], getslice=[lambda a, b, c: None], gt=[lambda a, b: None], iadd=[lambda a, b: None], iand=[lambda a, b: None], iconcat=[lambda a, b: None], idiv=[lambda a, b: None], ifloordiv=[lambda a, b: None], ilshift=[lambda a, b: None], imatmul=[lambda a, b: None], imod=[lambda a, b: None], imul=[lambda a, b: None], index=[lambda a: None], indexOf=[lambda a, b: None], inv=[lambda a: None], invert=[lambda a: None], ior=[lambda a, b: None], ipow=[lambda a, b: None], irepeat=[lambda a, b: None], irshift=[lambda a, b: None], is_=[lambda a, b: None], is_not=[lambda a, b: None], isCallable=[lambda a: None], isMappingType=[lambda a: None], isNumberType=[lambda a: None], isSequenceType=[lambda a: None], isub=[lambda a, b: None], itemgetter=[lambda item, *args: None], itruediv=[lambda a, b: None], ixor=[lambda a, b: None], le=[lambda a, b: None], length_hint=[lambda obj: None, lambda obj, default: None], lshift=[lambda a, b: None], lt=[lambda a, b: None], matmul=[lambda a, b: None], methodcaller=[lambda name, *args, **kwargs: None], mod=[lambda a, b: None], mul=[lambda a, b: None], ne=[lambda a, b: None], neg=[lambda a: None], not_=[lambda a: None], or_=[lambda a, b: None], pos=[lambda a: None], pow=[lambda a, b: None], repeat=[lambda a, b: None], rshift=[lambda a, b: None], sequenceIncludes=[lambda a, b: None], setitem=[lambda a, b, c: None], setslice=[lambda a, b, c, d: None], sub=[lambda a, b: None], truediv=[lambda a, b: None], truth=[lambda a: None], xor=[lambda a, b: None])
module_info['humpy_toolz'] = dict(curry=[(0, lambda *args, **kwargs: None)], excepts=[(0, lambda exc, func, handler=None: None)], flip=[(0, lambda func=None, a=None, b=None: None)], juxt=[(0, lambda *funcs: None)], memoize=[(0, lambda func=None, cache=None, key=None: None)])
module_info['humpy_toolz.functoolz'] = dict(Compose=[(0, lambda funcs: None)], InstanceProperty=[(0, lambda fget=None, fset=None, fdel=None, doc=None, classval=None: None)])

def num_pos_args(sigspec):
    """ Return the number of positional arguments.  ``f(x, y=1)`` has 1"""
    return sum((1 for x in sigspec.parameters.values() if x.kind == x.POSITIONAL_OR_KEYWORD and x.default is x.empty))

def get_exclude_keywords(num_pos_only, sigspec):
    """ Return the names of position-only arguments if func has **kwargs"""
    if num_pos_only == 0:
        return ()
    has_kwargs = any((x.kind == x.VAR_KEYWORD for x in sigspec.parameters.values()))
    if not has_kwargs:
        return ()
    pos_args = list(sigspec.parameters.values())[:num_pos_only]
    return tuple((x.name for x in pos_args))

def signature_or_spec(func):
    try:
        return inspect.signature(func)
    except (ValueError, TypeError):
        return None

def expand_sig(sig):
    """ Convert the signature spec in ``module_info`` to add to ``signatures``

    The input signature spec is one of:
        - ``lambda_func``
        - ``(num_position_args, lambda_func)``
        - ``(num_position_args, lambda_func, keyword_only_args)``

    The output signature spec is:
        ``(num_position_args, lambda_func, keyword_exclude, sigspec)``

    where ``keyword_exclude`` includes keyword only arguments and, if variadic
    keywords is present, the names of position-only argument.  The latter is
    included to support builtins such as ``partial(func, *args, **kwargs)``,
    which allows ``func=`` to be used as a keyword even though it's the name
    of a positional argument.
    """
    if isinstance(sig, tuple):
        if len(sig) == 3:
            num_pos_only, func, keyword_only = sig
            assert isinstance(sig[-1], tuple)
        else:
            num_pos_only, func = sig
            keyword_only = ()
        sigspec = signature_or_spec(func)
    else:
        func = sig
        sigspec = signature_or_spec(func)
        num_pos_only = num_pos_args(sigspec)
        keyword_only = ()
    keyword_exclude = get_exclude_keywords(num_pos_only, sigspec)
    return (num_pos_only, func, keyword_only + keyword_exclude, sigspec)
signatures = {}

def create_signature_registry(module_info=module_info, signatures=signatures):
    for module, info in module_info.items():
        if isinstance(module, str):
            module = import_module(module)
        for name, sigs in info.items():
            if hasattr(module, name):
                new_sigs = tuple((expand_sig(sig) for sig in sigs))
                signatures[getattr(module, name)] = new_sigs

def check_valid(sig, args, kwargs):
    """ Like ``is_valid_args`` for the given signature spec"""
    num_pos_only, func, keyword_exclude, sigspec = sig
    if len(args) < num_pos_only:
        return False
    if keyword_exclude:
        kwargs = dict(kwargs)
        for item in keyword_exclude:
            kwargs.pop(item, None)
    try:
        func(*args, **kwargs)
        return True
    except TypeError:
        return False

def _is_valid_args(func, args, kwargs):
    """ Like ``is_valid_args`` for builtins in our ``signatures`` registry"""
    if func not in signatures:
        return None
    sigs = signatures[func]
    return any((check_valid(sig, args, kwargs) for sig in sigs))

def check_partial(sig, args, kwargs):
    """ Like ``is_partial_args`` for the given signature spec"""
    num_pos_only, func, keyword_exclude, sigspec = sig
    if len(args) < num_pos_only:
        pad = (None,) * (num_pos_only - len(args))
        args = args + pad
    if keyword_exclude:
        kwargs = dict(kwargs)
        for item in keyword_exclude:
            kwargs.pop(item, None)
    return is_partial_args(func, args, kwargs, sigspec)

def _is_partial_args(func, args, kwargs):
    """ Like ``is_partial_args`` for builtins in our ``signatures`` registry"""
    if func not in signatures:
        return None
    sigs = signatures[func]
    return any((check_partial(sig, args, kwargs) for sig in sigs))

def check_arity(n, sig):
    num_pos_only, func, keyword_exclude, sigspec = sig
    if keyword_exclude or num_pos_only > n:
        return False
    return is_arity(n, func, sigspec)

def _is_arity(n, func):
    if func not in signatures:
        return None
    sigs = signatures[func]
    checks = [check_arity(n, sig) for sig in sigs]
    if all(checks):
        return True
    elif any(checks):
        return None
    return False

def check_varargs(sig):
    num_pos_only, func, keyword_exclude, sigspec = sig
    return has_varargs(func, sigspec)

def _has_varargs(func):
    if func not in signatures:
        return None
    sigs = signatures[func]
    checks = [check_varargs(sig) for sig in sigs]
    if all(checks):
        return True
    elif any(checks):
        return None
    return False

def check_keywords(sig):
    num_pos_only, func, keyword_exclude, sigspec = sig
    if keyword_exclude:
        return True
    return has_keywords(func, sigspec)

def _has_keywords(func):
    if func not in signatures:
        return None
    sigs = signatures[func]
    checks = [check_keywords(sig) for sig in sigs]
    if all(checks):
        return True
    elif any(checks):
        return None
    return False

def check_required_args(sig):
    num_pos_only, func, keyword_exclude, sigspec = sig
    return num_required_args(func, sigspec)

def _num_required_args(func):
    if func not in signatures:
        return None
    sigs = signatures[func]
    vals = [check_required_args(sig) for sig in sigs]
    val = vals[0]
    if all((x == val for x in vals)):
        return val
    return None



"""Tests for `Ordinals` ordering behaviors.

(AI generated docstring)

This module validates comparison semantics and sorting consistency across built-in and custom comparables using `pytest`.

"""
from dataclasses import dataclass
from hunterMakesPy import CallableFunction, Ordinals
from hunterMakesPy.tests.conftest import uniformTestFailureMessage
from typing import Self, TYPE_CHECKING, TypeVar
import functools
import inspect
import pytest

if TYPE_CHECKING:
	from collections.abc import Callable

小于 = TypeVar('小于', bound=Ordinals)
"""Type variable bound to `Ordinals`.

(AI generated docstring)

"""

def between(floor: 小于, ceiling: 小于, comparand: 小于) -> bool:
	"""Return whether `comparand` lies between `floor` and `ceiling` inclusive.

	Parameters
	----------
	floor : 小于
		Lower bound for `comparand`.
	ceiling : 小于
		Upper bound for `comparand`.
	comparand : 小于
		Value compared against `floor` and `ceiling`.

	Returns
	-------
	isBetween : bool
		Whether `comparand` is within the inclusive bounds.

	"""
	return floor <= comparand <= ceiling

@dataclass(frozen=True, slots=True)
class ComparableCardinal:
	"""Comparable wrapper for an `int` value.

	(AI generated docstring)

	Attributes
	----------
	value : int
		Integer used for ordering comparisons.

	"""

	value: int
	"""Integer used for ordering comparisons.

	(AI generated docstring)

	"""

	def __le__(self: Self, other: Self, /) -> bool:
		"""Return whether `self` is less than or equal to `other`.

		(AI generated docstring)

		Parameters
		----------
		self : Self
			Instance compared against `other`.
		other : Self
			Instance compared against `self`.

		Returns
		-------
		isLessThanOrEqual : bool
			Whether `self` is ordered before or equal to `other`.

		"""
		return self.value <= other.value

	def __lt__(self: Self, other: Self, /) -> bool:
		"""Return whether `self` is strictly less than `other`.

		(AI generated docstring)

		Parameters
		----------
		self : Self
			Instance compared against `other`.
		other : Self
			Instance compared against `self`.

		Returns
		-------
		isLessThan : bool
			Whether `self` is ordered before `other`.

		"""
		return self.value < other.value

@pytest.mark.parametrize(
	'floor, ceiling, comparand, expected'
	, [ (13, 34, 21, True) , (227, 695, 88, False) ]
)
def testOrdinalsBetweenWorksForInt(floor: int, ceiling: int, comparand: int, expected: bool) -> None:
	"""Verify `between` handles `int` operands.

	(AI generated docstring)

	Parameters
	----------
	floor : int
		Lower bound for `comparand`.
	ceiling : int
		Upper bound for `comparand`.
	comparand : int
		Value of `comparand` compared against the bounds.
	expected : bool
		Expected outcome from `between`.

	Returns
	-------
	unusedReturnValue : None
		Returns `None`.

	"""
	actual: bool = between(floor, ceiling, comparand)
	assert actual == expected, uniformTestFailureMessage( expected , actual , 'between' , floor , ceiling , comparand )

@pytest.mark.parametrize(
	'values, expected'
	, [ ([21, 13, 34, 8], [8, 13, 21, 34]) ]
)
def testOrdinalsSortingWorksForInt(values: list[int], expected: list[int]) -> None:
	"""Verify sorting uses ordinal ordering for `int` values.

	(AI generated docstring)

	Parameters
	----------
	values : list[int]
		Input `values` to sort.
	expected : list[int]
		Expected sorted output for `values`.

	Returns
	-------
	unusedReturnValue : None
		Returns `None`.

	"""
	actual: list[int] = sorted(values)
	assert actual == expected, uniformTestFailureMessage( expected , actual , 'sorted' , values )

@pytest.mark.parametrize(
	'floor, ceiling, comparand, expected'
	, [ ('fibonacci', 'prime', 'omega', True) , ('fibonacci', 'prime', 'tango', False) ]
)
def testOrdinalsBetweenWorksForStr(floor: str, ceiling: str, comparand: str, expected: bool) -> None:
	"""Verify `between` handles `str` operands.

	(AI generated docstring)

	Parameters
	----------
	floor : str
		Lower bound for `comparand`.
	ceiling : str
		Upper bound for `comparand`.
	comparand : str
		Value of `comparand` compared against the bounds.
	expected : bool
		Expected outcome from `between`.

	Returns
	-------
	unusedReturnValue : None
		Returns `None`.

	"""
	actual: bool = between(floor, ceiling, comparand)
	assert actual == expected, uniformTestFailureMessage( expected , actual , 'between' , floor , ceiling , comparand )

@pytest.mark.parametrize(
	'values, expected'
	, [ (['prime', 'fibonacci', 'omega'], ['fibonacci', 'omega', 'prime']) ]
)
def testOrdinalsSortingWorksForStr(values: list[str], expected: list[str]) -> None:
	"""Verify sorting uses ordinal ordering for `str` values.

	(AI generated docstring)

	Parameters
	----------
	values : list[str]
		Input `values` to sort.
	expected : list[str]
		Expected sorted output for `values`.

	Returns
	-------
	unusedReturnValue : None
		Returns `None`.

	"""
	actual: list[str] = sorted(values)
	assert actual == expected, uniformTestFailureMessage( expected , actual , 'sorted' , values )

@pytest.mark.parametrize(
	'floor, ceiling, comparand, expected'
	, [ ((13, 17), (21, 2), (13, 19), True) , ((13, 17), (21, 2), (8, 34), False) ]
)
def testOrdinalsBetweenWorksForTuple(floor: tuple[int, int], ceiling: tuple[int, int], comparand: tuple[int, int], expected: bool) -> None:
	"""Verify `between` handles `tuple[int, int]` operands.

	(AI generated docstring)

	Parameters
	----------
	floor : tuple[int, int]
		Lower bound for `comparand`.
	ceiling : tuple[int, int]
		Upper bound for `comparand`.
	comparand : tuple[int, int]
		Value of `comparand` compared against the bounds.
	expected : bool
		Expected outcome from `between`.

	Returns
	-------
	unusedReturnValue : None
		Returns `None`.

	"""
	actual: bool = between(floor, ceiling, comparand)
	assert actual == expected, uniformTestFailureMessage( expected , actual , 'between' , floor , ceiling , comparand )

@pytest.mark.parametrize( 'values, expected' , [ ( [(21, 2), (13, 19), (13, 17), (8, 34)] , [(8, 34), (13, 17), (13, 19), (21, 2)] ) ] )
def testOrdinalsSortingWorksForTuple(values: list[tuple[int, int]], expected: list[tuple[int, int]]) -> None:
	"""Verify sorting uses ordinal ordering for `tuple[int, int]` values.

	(AI generated docstring)

	Parameters
	----------
	values : list[tuple[int, int]]
		Input `values` to sort.
	expected : list[tuple[int, int]]
		Expected sorted output for `values`.

	Returns
	-------
	unusedReturnValue : None
		Returns `None`.

	"""
	actual: list[tuple[int, int]] = sorted(values)
	assert actual == expected, uniformTestFailureMessage( expected , actual , 'sorted' , values )

@pytest.mark.parametrize(
	'floor, ceiling, comparand, expected'
	, [
		(ComparableCardinal(13), ComparableCardinal(34), ComparableCardinal(21), True)
		, (ComparableCardinal(13), ComparableCardinal(34), ComparableCardinal(8), False)
	]
)
def testOrdinalsBetweenWorksForCustomComparable( floor: ComparableCardinal, ceiling: ComparableCardinal, comparand: ComparableCardinal, expected: bool ) -> None:
	"""Verify `between` handles `ComparableCardinal` operands.

	(AI generated docstring)

	Parameters
	----------
	floor : ComparableCardinal
		Lower `ComparableCardinal` bound for `comparand`.
	ceiling : ComparableCardinal
		Upper `ComparableCardinal` bound for `comparand`.
	comparand : ComparableCardinal
		Value of `comparand` compared against the bounds.
	expected : bool
		Expected outcome from `between`.

	Returns
	-------
	unusedReturnValue : None
		Returns `None`.

	"""
	actual: bool = between(floor, ceiling, comparand)
	assert actual == expected, uniformTestFailureMessage( expected , actual , 'between' , floor , ceiling , comparand )

@pytest.mark.parametrize( 'comparisonMethodName' , [ '__le__' , '__lt__' ] )
def testOrdinalsComparisonMethodsAcceptOtherOperand(comparisonMethodName: str) -> None:
	"""Validate `Ordinals` comparison method signatures.

	(AI generated docstring)

	Parameters
	----------
	comparisonMethodName : str
		Name of the `Ordinals` comparison method to inspect.

	Returns
	-------
	unusedReturnValue : None
		Returns `None`.

	"""
	comparisonMethod: Callable[[Ordinals, Ordinals], bool] = getattr(Ordinals, comparisonMethodName)
	signature: inspect.Signature = inspect.signature(comparisonMethod)
	listParameters: list[inspect.Parameter] = list(signature.parameters.values())

	assert len(listParameters) == 2, uniformTestFailureMessage(
		2 , len(listParameters) , 'Ordinals comparison parameter count' , comparisonMethodName , signature )

	listKinds: list[inspect._ParameterKind] = [parameter.kind for parameter in listParameters]
	expectedKinds: list[inspect._ParameterKind] = [inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.POSITIONAL_ONLY]
	assert listKinds == expectedKinds, uniformTestFailureMessage(
		expectedKinds , listKinds , 'Ordinals comparison parameter kinds' , comparisonMethodName , signature )

	actualReturnAnnotation: object = signature.return_annotation
	assert actualReturnAnnotation is bool, uniformTestFailureMessage(
		bool , actualReturnAnnotation , 'Ordinals comparison return annotation' , comparisonMethodName , signature )




from collections import defaultdict
from humpy_toolz.curried import first, merge, merge_with, operator as cop, reduce, second, sorted, take
from importlib import import_module
from operator import add
import humpy_toolz
import humpy_toolz.curried

def test_take() -> None:
    assert list(take(2)([1, 2, 3])) == [1, 2]

def test_first() -> None:
    assert first is humpy_toolz.itertoolz.first

def test_merge() -> None:
    assert merge(factory=lambda: defaultdict(int))({1: 1}) == {1: 1}
    assert merge({1: 1}) == {1: 1}
    assert merge({1: 1}, factory=lambda: defaultdict(int)) == {1: 1}

def test_merge_with() -> None:
    assert merge_with(sum)({1: 1}, {1: 2}) == {1: 3}

def test_merge_with_list() -> None:
    assert merge_with(sum, [{'a': 1}, {'a': 2}]) == {'a': 3}

def test_sorted() -> None:
    assert sorted(key=second)([(1, 2), (2, 1)]) == [(2, 1), (1, 2)]

def test_reduce() -> None:
    assert reduce(add)((1, 2, 3)) == 6

def test_module_name() -> None:
    assert humpy_toolz.curried.__name__ == 'humpy_toolz.curried'

def should_curry(func: object) -> bool:
    if not callable(func) or isinstance(func, humpy_toolz.curry):
        return False
    nargs = humpy_toolz.functoolz.num_required_args(func)
    if nargs is None or nargs > 1:
        return True
    return nargs == 1 and humpy_toolz.functoolz.has_keywords(func)

def test_curried_operator() -> None:
    import operator
    for k, v in vars(cop).items():
        if not callable(v):
            continue
        if not isinstance(v, humpy_toolz.curry):
            try:
                v(1)
            except TypeError:
                try:
                    v('x')
                except TypeError:
                    pass
                else:
                    continue
                raise AssertionError('humpy_toolz.curried.operator.%s is not curried!' % k)
        assert should_curry(getattr(operator, k)) == isinstance(v, humpy_toolz.curry), k
    assert len(set(vars(cop)) & {'add', 'sub', 'mul'}) == 3

def test_curried_namespace() -> None:
    exceptions = import_module('humpy_toolz.curried.exceptions')
    namespace = {}

    def curry_namespace(ns):
        return {name: humpy_toolz.curry(f) if should_curry(f) else f for name, f in ns.items() if '__' not in name}
    from_toolz = curry_namespace(vars(humpy_toolz))
    from_exceptions = curry_namespace(vars(exceptions))
    namespace.update(humpy_toolz.merge(from_toolz, from_exceptions))
    namespace = humpy_toolz.valfilter(callable, namespace)
    curried_namespace = humpy_toolz.valfilter(callable, humpy_toolz.curried.__dict__)
    if namespace != curried_namespace:
        missing = set(namespace) - set(curried_namespace)
        if missing:
            raise AssertionError('There are missing functions in humpy_toolz.curried:\n    %s' % '    \n'.join(sorted(missing)))
        extra = set(curried_namespace) - set(namespace)
        if extra:
            raise AssertionError('There are extra functions in humpy_toolz.curried:\n    %s' % '    \n'.join(sorted(extra)))
        unequal = humpy_toolz.merge_with(list, namespace, curried_namespace)
        unequal = humpy_toolz.valfilter(lambda x: x[0] != x[1], unequal)
        messages = []
        for name, (orig_func, auto_func) in sorted(unequal.items()):
            if name in from_exceptions:
                messages.append('%s should come from humpy_toolz.curried.exceptions' % name)
            elif should_curry(getattr(humpy_toolz, name)):
                messages.append('%s should be curried from humpy_toolz' % name)
            else:
                messages.append('%s should come from humpy_toolz and NOT be curried' % name)
        raise AssertionError('\n'.join(messages))



from humpy_toolz.itertoolz import partition_all
from humpy_toolz.utils import no_default
import functools

def _reduce(func, seq, initial=None):
	if initial is None:
		return functools.reduce(func, seq)
	else:
		return functools.reduce(func, seq, initial)

def fold(binop, seq, default=no_default, map=map, chunksize=128, combine=None):
	"""
	Reduce without guarantee of ordered reduction.

	Parameters
	----------
	binops
		Associative operator. The associative property allows us to
		leverage a parallel map to perform reductions in parallel.

	inputs:

	``binop``     - associative operator. The associative property allows us to
					leverage a parallel map to perform reductions in parallel.

	``seq``       - a sequence to be aggregated
	``default``   - an identity element like 0 for ``add`` or 1 for mul

	``map``       - an implementation of ``map``. This may be parallel and
					determines how work is distributed.
	``chunksize`` - Number of elements of ``seq`` that should be handled
					within a single function call
	``combine``   - Binary operator to combine two intermediate results.
					If ``binop`` is of type (total, item) -> total
					then ``combine`` is of type (total, total) -> total
					Defaults to ``binop`` for common case of operators like add

	Fold chunks up the collection into blocks of size ``chunksize`` and then
	feeds each of these to calls to ``reduce``. This work is distributed
	with a call to ``map``, gathered back and then refolded to finish the
	computation. In this way ``fold`` specifies only how to chunk up data but
	leaves the distribution of this work to an externally provided ``map``
	function. This function can be sequential or rely on multithreading,
	multiprocessing, or even distributed solutions.

	If ``map`` intends to serialize functions it should be prepared to accept
	and serialize lambdas. Note that the standard ``pickle`` module fails
	here.

	Example
	-------

	>>> # Provide a parallel map to accomplish a parallel sum
	>>> from operator import add
	>>> fold(add, [1, 2, 3, 4], chunksize=2, map=map)
	10
	"""
	assert chunksize > 1
	if combine is None:
		combine = binop
	chunks = partition_all(chunksize, seq)
	if default == no_default:
		results = map(functools.partial(_reduce, binop), chunks)
	else:
		results = map(functools.partial(_reduce, binop, initial=default), chunks)
	results = list(results)
	if len(results) == 1:
		return results[0]
	else:
		return fold(combine, results, map=map, chunksize=chunksize)
